# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The godeps plugin can be used for godeps-enabled go projects.

These projects have a file containing information about the project's
dependencies. This file is typically called "dependencies.tsv," but may be
named anything.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - go-channel:
      (string, default: latest/stable)
      The Snap Store channel to install go from. If set to an empty string,
      go will be installed using the system's traditional package manager.

    - go-packages:
      (list of strings)
      Go packages to build/install, these must be a "main" package.
      Dependencies should have already been retrieved by the `godeps-file`
      used for this part.
      Packages that are not "main" will not cause an error, but would
      not be useful either.

    - godeps-file:
      (string)
      Path to the godeps dependencies file contained within the source
      (default: dependencies.tsv)

    - go-importpath:
      (string)
      This entry tells the checked out `source` to live within a certain path
      within `GOPATH`. This is required in order to work with absolute imports
      and import path checking.
"""

import logging
import os
import shutil

import snapcraft
from snapcraft import common
from snapcraft.internal import errors


logger = logging.getLogger(__name__)


class GodepsPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["go-channel"] = {
            "type": "string",
            "default": "latest/stable",
        }
        schema["properties"]["godeps-file"] = {
            "type": "string",
            "default": "dependencies.tsv",
        }
        schema["properties"]["go-importpath"] = {"type": "string"}
        schema["properties"]["go-packages"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        # The import path must be specified.
        schema["required"] = ["source", "go-importpath"]

        return schema

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["go-packages"]

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["go-channel", "godeps-file", "go-importpath"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._setup_base_tools(options.go_channel, project.info.get_build_base())

        self._gopath = os.path.join(self.partdir, "go")
        self._gopath_src = os.path.join(self._gopath, "src")
        self._gopath_bin = os.path.join(self._gopath, "bin")
        self._gopath_pkg = os.path.join(self._gopath, "pkg")

    def _setup_base_tools(self, go_channel, base):
        if go_channel:
            self.build_snaps.append("go/{}".format(go_channel))
        elif base in ("core", "core16", "core18"):
            self.build_packages.append("golang-go")
        else:
            raise errors.PluginBaseError(part_name=self.name, base=base)

    def pull(self):
        super().pull()

        path_in_gopath = os.path.join(self._gopath_src, self.options.go_importpath)
        os.makedirs(os.path.dirname(path_in_gopath), exist_ok=True)
        if os.path.islink(path_in_gopath):
            os.unlink(path_in_gopath)
        os.symlink(self.sourcedir, path_in_gopath)

        # Fetch and run godeps
        logger.info("Fetching godeps...")
        self._run(["go", "get", "github.com/rogpeppe/godeps"])

        logger.info("Obtaining project dependencies...")
        self._run(
            [
                "godeps",
                "-t",
                "-u",
                os.path.join(self.sourcedir, self.options.godeps_file),
            ]
        )

    def clean_pull(self):
        super().clean_pull()

        # Remove the gopath (if present)
        if os.path.exists(self._gopath):
            shutil.rmtree(self._gopath)

    def build(self):
        super().build()

        for go_package in self.options.go_packages:
            self._run(["go", "install", go_package])
        if not self.options.go_packages:
            self._run(["go", "install", "./{}/...".format(self.options.go_importpath)])

        install_bin_path = os.path.join(self.installdir, "bin")
        os.makedirs(install_bin_path, exist_ok=True)
        os.makedirs(self._gopath_bin, exist_ok=True)
        for binary in os.listdir(self._gopath_bin):
            # Skip godeps. It serves no purpose in production.
            if binary == "godeps":
                continue

            binary_path = os.path.join(self._gopath_bin, binary)
            shutil.copy2(binary_path, install_bin_path)

    def clean_build(self):
        super().clean_build()

        if os.path.isdir(self._gopath_bin):
            shutil.rmtree(self._gopath_bin)

        if os.path.isdir(self._gopath_pkg):
            shutil.rmtree(self._gopath_pkg)

    def _run(self, cmd, **kwargs):
        env = self._build_environment()
        return self.run(cmd, cwd=self._gopath_src, env=env, **kwargs)

    def _build_environment(self):
        env = os.environ.copy()
        env["GOPATH"] = self._gopath
        env["GOBIN"] = self._gopath_bin

        # Add $GOPATH/bin so godeps is actually callable.
        env["PATH"] = "{}:{}".format(
            os.path.join(self._gopath, "bin"), env.get("PATH", "")
        )

        include_paths = []
        for root in [self.installdir, self.project.stage_dir]:
            include_paths.extend(
                common.get_library_paths(root, self.project.arch_triplet)
            )

        flags = common.combine_paths(include_paths, "-L", " ")
        env["CGO_LDFLAGS"] = "{} {} {}".format(
            env.get("CGO_LDFLAGS", ""), flags, env.get("LDFLAGS", "")
        )

        return env



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING

from . import errors
from ._lxd import LXD
from ._multipass import Multipass

if TYPE_CHECKING:
    from typing import Type  # noqa: F401

    from ._base_provider import Provider  # noqa: F401


def get_provider_for(provider_name: str) -> "Type[Provider]":
    """Returns a Type that can build with provider_name."""
    if provider_name == "multipass":
        return Multipass
    elif provider_name == "lxd":
        return LXD
    else:
        raise errors.ProviderNotSupportedError(provider=provider_name)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Any, Dict
from typing import Tuple

from .. import errors


class Extension(metaclass=abc.ABCMeta):
    """Extension is the class from which all extensions inherit.

    Extensions have the ability to add snippets to apps, parts, and indeed add new parts
    to a given snapcraft.yaml. All they need to do is define the proper variables.

    :ivar root_snippet: Instance variable, dict of properties to apply to root of the
                        snapcraft.yaml.
    :ivar app_snippet: Instance variable, dict of properties to apply to apps using this
                       extension.
    :ivar part_snippet: Instance variable, dict of properties to apply to parts using
                        this extension.
    :ivar parts: Instance variable, dict of part definitions required by this extension.
    """

    @staticmethod
    @abc.abstractmethod
    def get_supported_bases() -> Tuple[str, ...]:
        """Return a tuple of supported bases."""

    @staticmethod
    @abc.abstractmethod
    def get_supported_confinement() -> Tuple[str, ...]:
        """Return a tuple of supported confinement settings."""

    def __init__(self, *, extension_name: str, yaml_data: Dict[str, Any]) -> None:
        """Create a new Extension.

        :param str extension_name: The name of the extension.
        :param dict yaml_data: Loaded snapcraft.yaml data.
        """
        self._sanity_check(extension_name=extension_name, yaml_data=yaml_data)

        self.root_snippet = dict()  # type: Dict[str, Any]
        self.app_snippet = dict()  # type: Dict[str, Any]
        self.part_snippet = dict()  # type: Dict[str, Any]
        self.parts = dict()  # type: Dict[str, Any]

    def _sanity_check(self, *, extension_name: str, yaml_data: Dict[str, Any]) -> None:
        base = yaml_data.get("base")

        # A base is required in order to use extensions, so raise an error if not specified.
        if not base:
            raise errors.ExtensionBaseRequiredError()

        if base not in self.get_supported_bases():
            raise errors.ExtensionUnsupportedBaseError(extension_name, base)

        # Default to devmode if confinement is not set.
        confinement = yaml_data.get("confinement", "devmode")
        if confinement not in self.get_supported_confinement():
            raise errors.ExtensionUnsupportedConfinementError(
                extension_name, confinement
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import yaml
from typing import Any, Dict, TextIO, Union

try:
    # The C-based loaders/dumpers aren't available everywhere, but they're much faster.
    # Use them if possible. If not, we could fallback to the normal loader/dumper, but
    # they actually behave differently, so raise an error instead.
    from yaml import CSafeLoader, CSafeDumper  # type: ignore
except ImportError:
    raise RuntimeError("Snapcraft requires PyYAML to be built with libyaml bindings")


def load(stream: TextIO) -> Any:
    """Safely load YAML in ordered manner."""
    return yaml.load(stream, Loader=_SafeOrderedLoader)


def dump(data: Union[Dict[str, Any], yaml.YAMLObject], *, stream: TextIO = None) -> str:
    """Safely dump YAML in ordered manner."""
    return yaml.dump(
        data, stream, _SafeOrderedDumper, default_flow_style=False, allow_unicode=True
    )


class _SafeOrderedLoader(CSafeLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, _dict_constructor
        )


class _SafeOrderedDumper(CSafeDumper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_representer(str, _str_presenter)
        self.add_representer(collections.OrderedDict, _dict_representer)


class SnapcraftYAMLObject(yaml.YAMLObject):
    yaml_loader = _SafeOrderedLoader
    yaml_dumper = _SafeOrderedDumper

    # We could implement a from_yaml class method here which would force loading to
    # go through constructors, but our previous method of doing that has been broken
    # long enough that it would require a larger change.


def _dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


def _dict_constructor(loader, node):
    # Necessary in order to make yaml merge tags work
    loader.flatten_mapping(node)
    value = loader.construct_pairs(node)

    try:
        return collections.OrderedDict(value)
    except TypeError:
        raise yaml.constructor.ConstructorError(
            "while constructing a mapping",
            node.start_mark,
            "found unhashable key",
            node.start_mark,
        )


def _str_presenter(dumper, data):
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


class OctInt(SnapcraftYAMLObject):
    """An int represented in octal form."""

    yaml_tag = u"!OctInt"

    def __init__(self, value):
        super().__init__()
        self._value = value

    @classmethod
    def to_yaml(cls, dumper, data):
        """
        Convert a Python object to a representation node.
        """
        return dumper.represent_scalar(
            "tag:yaml.org,2002:int", "{:04o}".format(data._value)
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from datetime import datetime

from snapcraft.internal.deprecations import handle_deprecation_notice
from ._project_options import ProjectOptions
from ._project_info import ProjectInfo  # noqa: F401


class Project(ProjectOptions):
    """All details around building a project concerning the build environment
    and the snap being built."""

    def __init__(
        self,
        *,
        target_deb_arch=None,
        debug=False,
        snapcraft_yaml_file_path=None,
        work_dir: str = None,
        is_managed_host: bool = False
    ) -> None:

        project_dir = os.getcwd()
        if is_managed_host:
            work_dir = os.path.expanduser("~")
        else:
            work_dir = project_dir

        # This here check is mostly for backwards compatibility with the
        # rest of the code base.
        if snapcraft_yaml_file_path is None:
            self.info = None  # type: ProjectInfo
        else:
            self.info = ProjectInfo(snapcraft_yaml_file_path=snapcraft_yaml_file_path)

        self._is_managed_host = is_managed_host
        self._project_dir = project_dir
        self._work_dir = work_dir

        super().__init__(target_deb_arch, debug, work_dir=work_dir)

        self.local_plugins_dir = self._get_local_plugins_dir()
        self._start_time = datetime.utcnow()

    def _get_snapcraft_assets_dir(self) -> str:
        # Many test cases don't set the yaml file path and assume the default dir
        if not self.info:
            return os.path.join(self._project_dir, "snap")

        if self.info.snapcraft_yaml_file_path.endswith(
            os.path.join("build-aux", "snap", "snapcraft.yaml")
        ):
            return os.path.join(self._project_dir, "build-aux", "snap")
        else:
            return os.path.join(self._project_dir, "snap")

    def _get_local_plugins_dir(self) -> str:
        deprecated_plugins_dir = os.path.join(self._parts_dir, "plugins")
        if os.path.exists(deprecated_plugins_dir):
            handle_deprecation_notice("dn2")
            return deprecated_plugins_dir
        else:
            assets_dir = self._get_snapcraft_assets_dir()
            return os.path.join(assets_dir, "plugins")

    def _get_global_state_file_path(self) -> str:
        if self._is_managed_host:
            state_file_path = os.path.join(self._work_dir, "state")
        else:
            state_file_path = os.path.join(self._parts_dir, ".snapcraft_global_state")

        return state_file_path

    def _get_start_time(self) -> datetime:
        """Returns the timestamp for when a snapcraft project was loaded."""
        return self._start_time



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This plugin just dumps the content from a specified source.

This plugin uses the common plugin keywords as well as those for 'sources'.
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

In the cases where dumping the content needs some mangling or organizing
one would take advantage of the core functionalities available to plugins
such as: `filesets`, `stage`, `snap` and `organize`.
"""

import os

import snapcraft
from snapcraft.internal import errors


class DumpInvalidSymlinkError(errors.SnapcraftError):
    fmt = (
        "Failed to copy {path!r}: it's a symlink pointing outside the snap.\n"
        "Fix it to be valid when snapped and try again."
    )

    def __init__(self, path):
        super().__init__(path=path)


class DumpPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["required"] = ["source"]
        return schema

    def enable_cross_compilation(self):
        pass

    def build(self):
        super().build()
        snapcraft.file_utils.link_or_copy_tree(
            self.builddir,
            self.installdir,
            copy_function=lambda src, dst: _link_or_copy(src, dst, self.installdir),
        )


def _link_or_copy(source, destination, boundary):
    """Attempt to copy symlinks as symlinks unless pointing out of boundary."""

    follow_symlinks = False

    # If this is a symlink, analyze where it's pointing and make sure it will
    # still be valid when snapped. If it won't, follow the symlink when
    # copying (i.e. copy the file to which the symlink is pointing instead).
    if os.path.islink(source):
        link = os.readlink(source)
        destination_dirname = os.path.dirname(destination)
        normalized = os.path.normpath(os.path.join(destination_dirname, link))
        if os.path.isabs(link) or not normalized.startswith(boundary):
            # Only follow symlinks that are NOT pointing at libc (LP: #1658774)
            if link not in snapcraft.repo.Repo.get_package_libraries("libc6"):
                follow_symlinks = True

    try:
        snapcraft.file_utils.link_or_copy(
            source, destination, follow_symlinks=follow_symlinks
        )
    except errors.SnapcraftCopyFileNotFoundError:
        raise DumpInvalidSymlinkError(source)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import snapcraft.internal.errors
import os.path
import site
import sys


def _find_windows_data_dir(topdir):
    # On Windows we need to search for data directory:
    #
    # Option (a) - Running with pip install --user.
    # > topdir=C:\Users\chris\AppData\Roaming\Python\Python37\site-packages
    # > site.USER_BASE=C:\Users\chris\AppData\Roaming\Python
    # > sys.prefix=c:\program files\python37
    #
    # In this case, data_dir should be:
    # > C:\Users\chris\AppData\Roaming\Python\share\snapcraft
    #
    # Option (b) - Running with pyinstaller frozen exe:
    # > topdir=C:\Users\chris\AppData\Local\Temp\_MEI110562
    # > site.USER_BASE=
    # > sys.prefix=C:\Users\chris\AppData\Local\Temp\_MEI110562
    #
    # In this case, data_dir should be:
    # > C:\Users\chris\AppData\Local\Temp\_MEI110562\share\snapcraft
    #
    # Option (c) - Running with pip install (system-wide).
    # > topdir=c:\program files\python37\lib\site-packages
    # > site.USER_BASE=C:\Users\chris\AppData\Roaming\Python
    # > sys.prefix=c:\program files\python37
    #
    # In this case, data_dir should be:
    # > c:\program files\python37\share\snapcraft

    # Handle Option (a).
    if topdir.startswith(site.USER_BASE):
        data_dir = os.path.join(site.USER_BASE, "share", "snapcraft")
        if os.path.exists(data_dir):
            return data_dir

    # Handle Options (b) & (c).
    if topdir.startswith(sys.prefix):
        data_dir = os.path.join(sys.prefix, "share", "snapcraft")
        if os.path.exists(data_dir):
            return data_dir

    # Do our best to find something...
    for data_dir in [topdir, sys.prefix, site.USER_BASE]:
        data_dir = os.path.join(data_dir, "share", "snapcraft")
        if os.path.exists(data_dir):
            return data_dir

    raise snapcraft.internal.errors.SnapcraftDataDirectoryMissingError()


def setup_dirs() -> None:
    """
    Ensure that snapcraft.common plugindir is setup correctly
    and support running out of a development snapshot
    """
    from snapcraft.internal import common

    topdir = os.path.abspath(os.path.join(__file__, "..", "..", ".."))

    # Only change the default if we are running from a checkout or from the
    # snap, or in Windows.
    if os.path.exists(os.path.join(topdir, "setup.py")):
        common.set_plugindir(os.path.join(topdir, "snapcraft", "plugins"))
        common.set_schemadir(os.path.join(topdir, "schema"))
        common.set_extensionsdir(os.path.join(topdir, "extensions"))
        common.set_keyringsdir(os.path.join(topdir, "keyrings"))

    # The default paths are relative to sys.prefix, which works well for
    # Snapcraft as a deb or in a venv. However, the Python plugin installs
    # packages into $SNAP/ as a prefix, while Python itself is contained in
    # $SNAP/usr/. As a result, using sys.prefix (which is '/usr') to find these
    # files won't work in the snap.
    elif common.is_snap():
        parent_dir = os.path.join(os.environ.get("SNAP"), "share", "snapcraft")
        common.set_plugindir(os.path.join(parent_dir, "plugins"))
        common.set_schemadir(os.path.join(parent_dir, "schema"))
        common.set_extensionsdir(os.path.join(parent_dir, "extensions"))
        common.set_keyringsdir(os.path.join(parent_dir, "keyrings"))
        common.set_legacy_snapcraft_dir(
            os.path.join(os.environ.get("SNAP"), "legacy_snapcraft")
        )

    elif sys.platform == "win32":
        common.set_plugindir(os.path.join(topdir, "snapcraft", "plugins"))

        data_dir = _find_windows_data_dir(topdir)
        common.set_schemadir(os.path.join(data_dir, "schema"))
        common.set_extensionsdir(os.path.join(data_dir, "extensions"))
        common.set_keyringsdir(os.path.join(data_dir, "keyrings"))

    else:
        # Make sure required data directories exist in the default locations.
        # Plugins and legacy snapcraft directory are not required.
        for d in [
            common.get_schemadir(),
            common.get_extensionsdir(),
            common.get_keyringsdir(),
        ]:
            if not os.path.exists(d):
                raise snapcraft.internal.errors.SnapcraftDataDirectoryMissingError()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import typing
import os

import click

from . import echo
from ._command import SnapcraftProjectCommand
from ._options import add_build_options, get_build_environment, get_project
from snapcraft.internal import (
    build_providers,
    deprecations,
    errors,
    lifecycle,
    project_loader,
    steps,
)
from snapcraft.project._sanity_checks import conduct_project_sanity_check
from ._errors import TRACEBACK_MANAGED, TRACEBACK_HOST


logger = logging.getLogger(__name__)


if typing.TYPE_CHECKING:
    from snapcraft.internal.project import Project  # noqa: F401


# TODO: when snap is a real step we can simplify the arguments here.
def _execute(  # noqa: C901
    step: steps.Step,
    parts: str,
    pack_project: bool = False,
    output: str = None,
    shell: bool = False,
    shell_after: bool = False,
    setup_prime_try: bool = False,
    **kwargs
) -> "Project":
    # Cleanup any previous errors.
    _clean_provider_error()

    build_environment = get_build_environment(**kwargs)
    project = get_project(is_managed_host=build_environment.is_managed_host, **kwargs)

    conduct_project_sanity_check(project)

    if build_environment.is_managed_host or build_environment.is_host:
        project_config = project_loader.load_config(project)
        lifecycle.execute(step, project_config, parts)
        if pack_project:
            _pack(project.prime_dir, output=output)
    else:
        build_provider_class = build_providers.get_provider_for(
            build_environment.provider
        )
        try:
            build_provider_class.ensure_provider()
        except build_providers.errors.ProviderNotFound as provider_error:
            if provider_error.prompt_installable:
                if echo.is_tty_connected() and echo.confirm(
                    "Support for {!r} needs to be set up. "
                    "Would you like to do that it now?".format(provider_error.provider)
                ):
                    build_provider_class.setup_provider(echoer=echo)
                else:
                    raise provider_error
            else:
                raise provider_error

        with build_provider_class(project=project, echoer=echo) as instance:
            instance.mount_project()
            try:
                if shell:
                    # shell means we want to do everything right up to the previous
                    # step and then go into a shell instead of the requested step.
                    # the "snap" target is a special snowflake that has not made its
                    # way to be a proper step.
                    previous_step = None
                    if pack_project:
                        previous_step = steps.PRIME
                    elif step > steps.PULL:
                        previous_step = step.previous_step()
                    # steps.PULL is the first step, so we would directly shell into it.
                    if previous_step:
                        instance.execute_step(previous_step)
                elif pack_project:
                    instance.pack_project(output=output)
                elif setup_prime_try:
                    instance.expose_prime()
                    instance.execute_step(step)
                else:
                    instance.execute_step(step)
            except Exception:
                _retrieve_provider_error(instance)
                if project.debug:
                    instance.shell()
                else:
                    echo.warning(
                        "Run the same command again with --debug to shell into the environment "
                        "if you wish to introspect this failure."
                    )
                    raise
            else:
                if shell or shell_after:
                    instance.shell()
    return project


def _pack(directory: str, *, output: str) -> None:
    snap_name = lifecycle.pack(directory, output)
    echo.info("Snapped {}".format(snap_name))


def _clean_provider_error() -> None:
    if os.path.isfile(TRACEBACK_HOST):
        try:
            os.remove(TRACEBACK_HOST)
        except Exception as e:
            logger.debug("can't remove error file: {}", str(e))


def _retrieve_provider_error(instance) -> None:
    try:
        instance.pull_file(TRACEBACK_MANAGED, TRACEBACK_HOST, delete=True)
    except Exception as e:
        logger.debug("can't retrieve error file: {}", str(e))


@click.group()
@add_build_options()
@click.pass_context
def lifecyclecli(ctx, **kwargs):
    pass


@lifecyclecli.command()
def init():
    """Initialize a snapcraft project."""
    snapcraft_yaml_path = lifecycle.init()
    echo.info("Created {}.".format(snapcraft_yaml_path))
    echo.wrapped(
        "Go to https://docs.snapcraft.io/the-snapcraft-format/8337 for more "
        "information about the snapcraft.yaml format."
    )


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@click.pass_context
@add_build_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def pull(ctx, parts, **kwargs):
    """Download or retrieve artifacts defined for a part.

    \b
    Examples:
        snapcraft pull
        snapcraft pull my-part1 my-part2

    """
    _execute(steps.PULL, parts, **kwargs)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@add_build_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def build(parts, **kwargs):
    """Build artifacts defined for a part.

    \b
    Examples:
        snapcraft build
        snapcraft build my-part1 my-part2

    """
    _execute(steps.BUILD, parts, **kwargs)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@add_build_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def stage(parts, **kwargs):
    """Stage the part's built artifacts into the common staging area.

    \b
    Examples:
        snapcraft stage
        snapcraft stage my-part1 my-part2

    """
    _execute(steps.STAGE, parts, **kwargs)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@add_build_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def prime(parts, **kwargs):
    """Final copy and preparation for the snap.

    \b
    Examples:
        snapcraft prime
        snapcraft prime my-part1 my-part2

    """
    _execute(steps.PRIME, parts, **kwargs)


@lifecyclecli.command("try")
@add_build_options()
def try_command(**kwargs):
    """Try a snap on the host, priming if necessary.

    This feature only works on snap enabled systems.

    \b
    Examples:
        snapcraft try

    """
    project = _execute(steps.PRIME, [], setup_prime_try=True, **kwargs)
    # project.prime_dir here points to the on-host prime directory.
    echo.info("You can now run `snap try {}`.".format(project.prime_dir))


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@add_build_options()
@click.argument("directory", required=False)
@click.option("--output", "-o", help="path to the resulting snap.")
def snap(directory, output, **kwargs):
    """Create a snap.

    \b
    Examples:
        snapcraft snap
        snapcraft snap --output renamed-snap.snap

    If you want to snap a directory, you should use the pack command
    instead.
    """
    if directory:
        deprecations.handle_deprecation_notice("dn6")
        _pack(directory, output=output)
    else:
        _execute(steps.PRIME, parts=[], pack_project=True, output=output, **kwargs)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@click.argument("directory")
@click.option("--output", "-o", help="path to the resulting snap.")
def pack(directory, output, **kwargs):
    """Create a snap from a directory holding a valid snap.

    The layout of <directory> should contain a valid meta/snap.yaml in
    order to be a valid snap.

    \b
    Examples:
        snapcraft pack my-snap-directory
        snapcraft pack my-snap-directory --output renamed-snap.snap

    """
    _pack(directory, output=output)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
@click.option(
    "--use-lxd",
    is_flag=True,
    required=False,
    help="Forces snapcraft to use LXD for this clean command.",
)
@click.option(
    "--destructive-mode",
    is_flag=True,
    required=False,
    help="Forces snapcraft to try and use the current host to clean.",
)
@click.option("--unprime", is_flag=True, required=False, hidden=True)
@click.option("--step", required=False, hidden=True)
def clean(parts, use_lxd, destructive_mode, unprime, step):
    """Remove a part's assets.

    \b
    Examples:
        snapcraft clean
        snapcraft clean my-part
    """
    # This option is only valid in legacy.
    if step:
        raise click.BadOptionUsage("--step", "no such option: --step")

    build_environment = get_build_environment(
        use_lxd=use_lxd, destructive_mode=destructive_mode
    )

    try:
        project = get_project(is_managed_host=build_environment.is_managed_host)
    except errors.ProjectNotFoundError:
        # Fresh environment, nothing to clean.
        return

    if unprime and not build_environment.is_managed_host:
        raise click.BadOptionUsage("--unprime", "no such option: --unprime")

    if build_environment.is_managed_host or build_environment.is_host:
        step = steps.PRIME if unprime else None
        lifecycle.clean(project, parts, step)
    else:
        build_provider_class = build_providers.get_provider_for(
            build_environment.provider
        )
        if parts:
            with build_provider_class(project=project, echoer=echo) as instance:
                instance.clean(part_names=parts)
        else:
            build_provider_class(project=project, echoer=echo).clean_project()
            # Clear the prime directory on the host
            lifecycle.clean(project, parts, steps.PRIME)


if __name__ == "__main__":
    lifecyclecli.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import getpass
import hashlib
import json
import logging
import operator
import os
import re
import subprocess
import tempfile
from datetime import datetime
from subprocess import Popen
from typing import Dict, Iterable, Optional, TextIO

# Ideally we would move stuff into more logical components
from snapcraft.cli import echo
from tabulate import tabulate

from snapcraft.file_utils import calculate_sha3_384, get_tool_path
from snapcraft import storeapi, yaml_utils
from snapcraft.internal import cache, deltas, repo
from snapcraft.internal.errors import SnapDataExtractionError, ToolMissingError
from snapcraft.internal.deltas.errors import (
    DeltaGenerationError,
    DeltaGenerationTooBigError,
)


logger = logging.getLogger(__name__)


def _get_data_from_snap_file(snap_path):
    with tempfile.TemporaryDirectory() as temp_dir:
        unsquashfs_path = get_tool_path("unsquashfs")
        try:
            output = subprocess.check_output(
                [
                    unsquashfs_path,
                    "-d",
                    os.path.join(temp_dir, "squashfs-root"),
                    snap_path,
                    "-e",
                    os.path.join("meta", "snap.yaml"),
                ]
            )
        except subprocess.CalledProcessError:
            raise SnapDataExtractionError(os.path.basename(snap_path))
        logger.debug(output)
        with open(
            os.path.join(temp_dir, "squashfs-root", "meta", "snap.yaml")
        ) as yaml_file:
            snap_yaml = yaml_utils.load(yaml_file)
    return snap_yaml


@contextlib.contextmanager
def _get_icon_from_snap_file(snap_path):
    icon_file = None
    with tempfile.TemporaryDirectory() as temp_dir:
        unsquashfs_path = get_tool_path("unsquashfs")
        try:
            output = subprocess.check_output(
                [
                    unsquashfs_path,
                    "-d",
                    os.path.join(temp_dir, "squashfs-root"),
                    snap_path,
                    "-e",
                    "meta/gui",
                ]
            )
        except subprocess.CalledProcessError:
            raise SnapDataExtractionError(os.path.basename(snap_path))
        logger.debug("Output extracting icon from snap: %s", output)
        for extension in ("png", "svg"):
            icon_name = "icon.{}".format(extension)
            icon_path = os.path.join(temp_dir, "squashfs-root", "meta/gui", icon_name)
            if os.path.exists(icon_path):
                icon_file = open(icon_path, "rb")
                break
        try:
            yield icon_file
        finally:
            if icon_file is not None:
                icon_file.close()


def _fail_login(msg: str = "") -> bool:
    echo.error(msg)
    echo.error("Login failed.")
    return False


def _get_url_from_error(error: storeapi.errors.StoreAccountInformationError) -> str:
    if error.extra:  # type: ignore
        return error.extra[0].get("url")  # type: ignore
    return ""


def _check_dev_agreement_and_namespace_statuses(store) -> None:
    """ Check the agreement and namespace statuses of the dev.
    Fail if either of those conditions is not met.
    Re-raise `StoreAccountInformationError` if we get an error and
    the error is not either of these.
    """
    # Check account information for the `developer agreement` status.
    try:
        store.get_account_information()
    except storeapi.errors.StoreAccountInformationError as e:
        if storeapi.constants.MISSING_AGREEMENT == e.error:  # type: ignore
            # A precaution if store does not return new style error.
            url = _get_url_from_error(e) or storeapi.constants.UBUNTU_STORE_TOS_URL
            choice = input(storeapi.constants.AGREEMENT_INPUT_MSG.format(url))
            if choice in {"y", "Y"}:
                try:
                    store.sign_developer_agreement(latest_tos_accepted=True)
                except:  # noqa LP: #1733003
                    raise storeapi.errors.NeedTermsSignedError(
                        storeapi.constants.AGREEMENT_SIGN_ERROR.format(url)
                    )
            else:
                raise storeapi.errors.NeedTermsSignedError(
                    storeapi.constants.AGREEMENT_ERROR
                )

    # Now check account information for the `namespace` status.
    try:
        store.get_account_information()
    except storeapi.errors.StoreAccountInformationError as e:
        if storeapi.constants.MISSING_NAMESPACE in e.error:  # type: ignore
            # A precaution if store does not return new style error.
            url = _get_url_from_error(e) or storeapi.constants.UBUNTU_STORE_ACCOUNT_URL
            raise storeapi.errors.NeedTermsSignedError(
                storeapi.constants.NAMESPACE_ERROR.format(url)
            )
        else:
            raise


def _try_login(
    email: str,
    password: str,
    *,
    store: storeapi.StoreClient = None,
    save: bool = True,
    packages: Iterable[Dict[str, str]] = None,
    acls: Iterable[str] = None,
    channels: Iterable[str] = None,
    expires: str = None,
    config_fd: TextIO = None
) -> None:
    try:
        store.login(
            email,
            password,
            packages=packages,
            acls=acls,
            channels=channels,
            expires=expires,
            config_fd=config_fd,
            save=save,
        )
        if not config_fd:
            print()
            logger.info(storeapi.constants.TWO_FACTOR_WARNING)
    except storeapi.errors.StoreTwoFactorAuthenticationRequired:
        one_time_password = input("Second-factor auth: ")
        store.login(
            email,
            password,
            one_time_password=one_time_password,
            acls=acls,
            packages=packages,
            channels=channels,
            expires=expires,
            config_fd=config_fd,
            save=save,
        )

    # Continue if agreement and namespace conditions are met.
    _check_dev_agreement_and_namespace_statuses(store)


def login(
    *,
    store: storeapi.StoreClient = None,
    packages: Iterable[Dict[str, str]] = None,
    save: bool = True,
    acls: Iterable[str] = None,
    channels: Iterable[str] = None,
    expires: str = None,
    config_fd: TextIO = None
) -> bool:
    if not store:
        store = storeapi.StoreClient()

    email = ""
    password = ""

    if not config_fd:
        print(
            "Enter your Ubuntu One e-mail address and password.\n"
            "If you do not have an Ubuntu One account, you can create one "
            "at https://snapcraft.io/account"
        )
        email = input("Email: ")
        if os.environ.get("SNAPCRAFT_TEST_INPUT"):
            password = input("Password: ")
        else:
            password = getpass.getpass("Password: ")

    try:
        _try_login(
            email,
            password,
            store=store,
            packages=packages,
            acls=acls,
            channels=channels,
            expires=expires,
            config_fd=config_fd,
            save=save,
        )
    # Let StoreAuthenticationError pass through so we get decent error messages
    except storeapi.errors.InvalidCredentialsError:
        return _fail_login(storeapi.constants.INVALID_CREDENTIALS)
    except storeapi.errors.StoreAccountInformationError:
        return _fail_login(storeapi.constants.ACCOUNT_INFORMATION_ERROR)
    except storeapi.errors.NeedTermsSignedError as e:
        return _fail_login(e.message)  # type: ignore

    return True


@contextlib.contextmanager
def _requires_login():
    try:
        yield
    except storeapi.errors.InvalidCredentialsError:
        logger.error("No valid credentials found." ' Have you run "snapcraft login"?')
        raise


def list_registered():
    series = storeapi.constants.DEFAULT_SERIES

    store = storeapi.StoreClient()
    with _requires_login():
        account_info = store.get_account_information()
    snaps = [
        (
            name,
            info["since"],
            "private" if info["private"] else "public",
            info["price"] or "-",
            "-",
        )
        for name, info in account_info["snaps"].get(series, {}).items()
        # Presenting only approved snap registrations, which means name
        # disputes will be displayed/sorted some other way.
        if info["status"] == "Approved"
    ]

    if not snaps:
        print("There are no registered snaps for series {!r}.".format(series))
        return

    tabulated_snaps = tabulate(
        sorted(snaps, key=operator.itemgetter(0)),
        headers=["Name", "Since", "Visibility", "Price", "Notes"],
        tablefmt="plain",
    )
    print(tabulated_snaps)


def _get_usable_keys(name=None):
    keys = json.loads(
        subprocess.check_output(["snap", "keys", "--json"], universal_newlines=True)
    )
    if keys is not None:
        for key in keys:
            if name is None or name == key["name"]:
                yield key


def _select_key(keys):
    if len(keys) > 1:
        print("Select a key:")
        print()
        tabulated_keys = tabulate(
            [(i + 1, key["name"], key["sha3-384"]) for i, key in enumerate(keys)],
            headers=["Number", "Name", "SHA3-384 fingerprint"],
            tablefmt="plain",
        )
        print(tabulated_keys)
        print()
        while True:
            try:
                keynum = int(input("Key number: ")) - 1
            except ValueError:
                continue
            if keynum >= 0 and keynum < len(keys):
                return keys[keynum]
    else:
        return keys[0]


def _export_key(name, account_id):
    return subprocess.check_output(
        ["snap", "export-key", "--account={}".format(account_id), name],
        universal_newlines=True,
    )


def list_keys():
    if not repo.Repo.is_package_installed("snapd"):
        raise storeapi.errors.MissingSnapdError("list-keys")
    keys = list(_get_usable_keys())
    store = storeapi.StoreClient()
    with _requires_login():
        account_info = store.get_account_information()
    enabled_keys = {
        account_key["public-key-sha3-384"]
        for account_key in account_info["account_keys"]
    }
    if enabled_keys:
        tabulated_keys = tabulate(
            [
                (
                    "*" if key["sha3-384"] in enabled_keys else "-",
                    key["name"],
                    key["sha3-384"],
                    "" if key["sha3-384"] in enabled_keys else "(not registered)",
                )
                for key in keys
            ],
            headers=["", "Name", "SHA3-384 fingerprint", ""],
            tablefmt="plain",
        )
        print(tabulated_keys)
    else:
        print(
            "No keys have been registered."
            " See 'snapcraft register-key --help' to register a key."
        )


def create_key(name):
    if not repo.Repo.is_package_installed("snapd"):
        raise storeapi.errors.MissingSnapdError("create-key")
    if not name:
        name = "default"
    keys = list(_get_usable_keys(name=name))
    if keys:
        # `snap create-key` would eventually fail, but we can save the user
        # some time in this obvious error case by not bothering to talk to
        # the store first.
        raise storeapi.errors.KeyAlreadyRegisteredError(name)
    store = storeapi.StoreClient()
    try:
        account_info = store.get_account_information()
        enabled_names = {
            account_key["name"] for account_key in account_info["account_keys"]
        }
    except storeapi.errors.InvalidCredentialsError:
        # Don't require a login here; if they don't have valid credentials,
        # then they probably also don't have a key registered with the store
        # yet.
        enabled_names = set()
    if name in enabled_names:
        raise storeapi.errors.KeyAlreadyRegisteredError(name)
    subprocess.check_call(["snap", "create-key", name])


def _maybe_prompt_for_key(name):
    keys = list(_get_usable_keys(name=name))
    if not keys:
        if name is not None:
            raise storeapi.errors.NoSuchKeyError(name)
        else:
            raise storeapi.errors.NoKeysError
    return _select_key(keys)


def register_key(name):
    if not repo.Repo.is_package_installed("snapd"):
        raise storeapi.errors.MissingSnapdError("register-key")
    key = _maybe_prompt_for_key(name)
    store = storeapi.StoreClient()
    try:
        if not login(store=store, acls=["modify_account_key"], save=False):
            raise storeapi.errors.LoginRequiredError()
    except storeapi.errors.StoreAuthenticationError as e:
        raise storeapi.errors.LoginRequiredError(str(e)) from e
    logger.info("Registering key ...")
    account_info = store.get_account_information()
    account_key_request = _export_key(key["name"], account_info["account_id"])
    store.register_key(account_key_request)
    logger.info(
        'Done. The key "{}" ({}) may be used to sign your assertions.'.format(
            key["name"], key["sha3-384"]
        )
    )


def register(snap_name: str, is_private: bool = False, store_id: str = None) -> None:
    logger.info("Registering {}.".format(snap_name))
    store = storeapi.StoreClient()
    with _requires_login():
        store.register(snap_name, is_private=is_private, store_id=store_id)


def _generate_snap_build(authority_id, snap_id, grade, key_name, snap_filename):
    """Return the signed snap-build declaration for a snap on disk."""
    cmd = [
        "snap",
        "sign-build",
        "--developer-id=" + authority_id,
        "--snap-id=" + snap_id,
        "--grade=" + grade,
    ]
    if key_name:
        cmd.extend(["-k", key_name])
    cmd.append(snap_filename)
    try:
        return subprocess.check_output(cmd)
    except subprocess.CalledProcessError as e:
        raise storeapi.errors.SignBuildAssertionError(snap_filename) from e


def sign_build(snap_filename, key_name=None, local=False):
    if not repo.Repo.is_package_installed("snapd"):
        raise storeapi.errors.MissingSnapdError("sign-build")

    if not os.path.exists(snap_filename):
        raise FileNotFoundError("The file {!r} does not exist.".format(snap_filename))

    snap_series = storeapi.constants.DEFAULT_SERIES
    snap_yaml = _get_data_from_snap_file(snap_filename)
    snap_name = snap_yaml["name"]
    grade = snap_yaml.get("grade", "stable")

    store = storeapi.StoreClient()
    with _requires_login():
        account_info = store.get_account_information()

    try:
        authority_id = account_info["account_id"]
        snap_id = account_info["snaps"][snap_series][snap_name]["snap-id"]
    except KeyError as e:
        raise storeapi.errors.StoreBuildAssertionPermissionError(
            snap_name, snap_series
        ) from e

    snap_build_path = snap_filename + "-build"
    if os.path.isfile(snap_build_path):
        logger.info("A signed build assertion for this snap already exists.")
        with open(snap_build_path, "rb") as fd:
            snap_build_content = fd.read()
    else:
        key = _maybe_prompt_for_key(key_name)
        if not local:
            is_registered = [
                a
                for a in account_info["account_keys"]
                if a["public-key-sha3-384"] == key["sha3-384"]
            ]
            if not is_registered:
                raise storeapi.errors.KeyNotRegisteredError(key["name"])
        snap_build_content = _generate_snap_build(
            authority_id, snap_id, grade, key["name"], snap_filename
        )
        with open(snap_build_path, "w+") as fd:
            fd.write(snap_build_content.decode())
        logger.info("Build assertion {} saved to disk.".format(snap_build_path))

    if not local:
        store.push_snap_build(snap_id, snap_build_content.decode())
        logger.info("Build assertion {} pushed to the Store.".format(snap_build_path))


def push_metadata(snap_filename, force):
    """Push only the metadata to the server.

    If force=True it will force the local metadata into the Store,
    ignoring any possible conflict.
    """
    logger.debug("Pushing metadata to the Store (force=%s)", force)

    # get the metadata from the snap
    snap_yaml = _get_data_from_snap_file(snap_filename)
    metadata = {
        "summary": snap_yaml["summary"],
        "description": snap_yaml["description"],
    }

    # other snap info
    snap_name = snap_yaml["name"]

    # hit the server
    store = storeapi.StoreClient()
    with _requires_login():
        store.push_precheck(snap_name)
        store.push_metadata(snap_name, metadata, force)
        with _get_icon_from_snap_file(snap_filename) as icon:
            metadata = {"icon": icon}
            store.push_binary_metadata(snap_name, metadata, force)

    logger.info("The metadata has been pushed")


def push(snap_filename, release_channels=None):
    """Push a snap_filename to the store.

    If a cached snap is available, a delta will be generated from
    the cached snap to the new target snap and uploaded instead. In the
    case of a delta processing or upload failure, push will fall back to
    uploading the full snap.

    If release_channels is defined it also releases it to those channels if the
    store deems the uploaded snap as ready to release.
    """
    snap_yaml = _get_data_from_snap_file(snap_filename)
    snap_name = snap_yaml["name"]
    built_at = snap_yaml.get("snapcraft-started-at")
    store = storeapi.StoreClient()

    logger.debug(
        "Run push precheck and verify cached data for {!r}.".format(snap_filename)
    )
    with _requires_login():
        store.push_precheck(snap_name)

    snap_cache = cache.SnapCache(project_name=snap_name)
    arch = "all"

    with contextlib.suppress(KeyError):
        arch = snap_yaml["architectures"][0]

    source_snap = snap_cache.get(deb_arch=arch)
    sha3_384_available = hasattr(hashlib, "sha3_384")

    if sha3_384_available and source_snap:
        try:
            result = _push_delta(snap_name, snap_filename, source_snap, built_at)
        except storeapi.errors.StoreDeltaApplicationError as e:
            logger.warning(
                "Error generating delta: {}\n"
                "Falling back to pushing full snap...".format(str(e))
            )
            result = _push_snap(snap_name, snap_filename, built_at)
        except storeapi.errors.StorePushError as e:
            store_error = e.error_list[0].get("message")
            logger.warning(
                "Unable to push delta to store: {}\n"
                "Falling back to pushing full snap...".format(store_error)
            )
            result = _push_snap(snap_name, snap_filename, built_at)
    else:
        result = _push_snap(snap_name, snap_filename, built_at)

    logger.info("Revision {!r} of {!r} created.".format(result["revision"], snap_name))

    snap_cache.cache(snap_filename=snap_filename)
    snap_cache.prune(deb_arch=arch, keep_hash=calculate_sha3_384(snap_filename))

    if release_channels:
        release(snap_name, result["revision"], release_channels)


def _push_snap(snap_name, snap_filename, built_at):
    store = storeapi.StoreClient()
    with _requires_login():
        tracker = store.upload(snap_name, snap_filename, built_at=built_at)
    result = tracker.track()
    tracker.raise_for_code()
    return result


def _push_delta(snap_name, snap_filename, source_snap, built_at):
    store = storeapi.StoreClient()
    delta_format = "xdelta3"
    logger.debug("Found cached source snap {}.".format(source_snap))
    target_snap = os.path.join(os.getcwd(), snap_filename)

    try:
        xdelta_generator = deltas.XDelta3Generator(
            source_path=source_snap, target_path=target_snap
        )
        delta_filename = xdelta_generator.make_delta()
    except (DeltaGenerationError, DeltaGenerationTooBigError, ToolMissingError) as e:
        raise storeapi.errors.StoreDeltaApplicationError(str(e))

    snap_hashes = {
        "source_hash": calculate_sha3_384(source_snap),
        "target_hash": calculate_sha3_384(target_snap),
        "delta_hash": calculate_sha3_384(delta_filename),
    }

    try:
        logger.debug("Pushing delta {!r}.".format(delta_filename))
        with _requires_login():
            delta_tracker = store.upload(
                snap_name,
                delta_filename,
                delta_format=delta_format,
                source_hash=snap_hashes["source_hash"],
                target_hash=snap_hashes["target_hash"],
                delta_hash=snap_hashes["delta_hash"],
                built_at=built_at,
            )
        result = delta_tracker.track()
        delta_tracker.raise_for_code()
    except storeapi.errors.StoreReviewError as e:
        if e.code == "processing_upload_delta_error":
            raise storeapi.errors.StoreDeltaApplicationError(str(e))
        else:
            raise
    except storeapi.errors.StoreServerError as e:
        raise storeapi.errors.StorePushError(snap_name, e.response)
    finally:
        if os.path.isfile(delta_filename):
            try:
                os.remove(delta_filename)
            except OSError:
                logger.warning("Unable to remove delta {}.".format(delta_filename))
    return result


def _get_text_for_opened_channels(opened_channels):
    if len(opened_channels) == 1:
        return "The {!r} channel is now open.".format(opened_channels[0])
    else:
        channels = ("{!r}".format(channel) for channel in opened_channels[:-1])
        return "The {} and {!r} channels are now open.".format(
            ", ".join(channels), opened_channels[-1]
        )


def _get_text_for_channel(channel):
    if channel["info"] == "none":
        channel_text = (channel["channel"], "-", "-", "")
    elif channel["info"] == "tracking":
        channel_text = (channel["channel"], "^", "^", "")
    elif channel["info"] == "specific":
        channel_text = (channel["channel"], channel["version"], channel["revision"], "")
    elif channel["info"] == "branch":
        channel_text = (
            channel["channel"],
            channel["version"],
            channel["revision"],
            channel["expires_at"],
        )
    else:
        logger.error(
            "Unexpected channel info: %r in channel %s",
            channel["info"],
            channel["channel"],
        )
        channel_text = (channel["channel"], "", "", "")

    return channel_text


def release(snap_name, revision, release_channels):
    store = storeapi.StoreClient()
    with _requires_login():
        channels = store.release(snap_name, revision, release_channels)
    channel_map_tree = channels.get("channel_map_tree", {})

    # This does not look good in green so we print instead
    tabulated_channels = _tabulated_channel_map_tree(channel_map_tree)
    print(tabulated_channels)

    if "opened_channels" in channels:
        logger.info(_get_text_for_opened_channels(channels["opened_channels"]))


def _tabulated_channel_map_tree(channel_map_tree):

    """Tabulate channel map (LTS Channel channel-maps)"""

    def _format_tree(channel_maps, track, series):
        arches = []

        for arch, channel_map in sorted(channel_maps.items()):
            arches += [
                (printable_arch,) + _get_text_for_channel(channel)
                for (printable_arch, channel) in zip(
                    [arch] + [""] * len(channel_map), channel_map
                )
            ]

        return [
            (printable_arch,) + printable_track
            for (printable_arch, printable_track) in zip(
                [track] + [""] * len(arches), arches
            )
        ]

    data = []
    for track, track_data in sorted(channel_map_tree.items()):
        channel_maps = {}
        for series, series_data in track_data.items():
            for arch, channel_map in series_data.items():
                channel_maps[arch] = channel_map
        parsed_channels = [
            channel for channel in _format_tree(channel_maps, track, series)
        ]
        data += parsed_channels

    have_expiration = any(x[5] for x in data)
    expires_at_header = "Expires at" if have_expiration else ""
    headers = ["Track", "Arch", "Channel", "Version", "Revision", expires_at_header]
    return tabulate(data, numalign="left", headers=headers, tablefmt="plain")


def close(snap_name, channel_names):
    """Close one or more channels for the specific snap."""
    snap_series = storeapi.constants.DEFAULT_SERIES

    store = storeapi.StoreClient()

    with _requires_login():
        info = store.get_account_information()

    try:
        snap_id = info["snaps"][snap_series][snap_name]["snap-id"]
    except KeyError as e:
        raise storeapi.errors.StoreChannelClosingPermissionError(
            snap_name, snap_series
        ) from e

    closed_channels, c_m_tree = store.close_channels(snap_id, channel_names)

    tabulated_status = _tabulated_channel_map_tree(c_m_tree)
    print(tabulated_status)

    print()
    if len(closed_channels) == 1:
        msg = "The {} channel is now closed.".format(closed_channels[0])
    else:
        msg = "The {} and {} channels are now closed.".format(
            ", ".join(closed_channels[:-1]), closed_channels[-1]
        )
    logger.info(msg)


def download(
    snap_name,
    *,
    arch: str,
    download_path: str,
    risk: str,
    track: Optional[str] = None,
    except_hash=""
):
    """Download snap from the store to download_path.
    :param str snap_name: The snap name to download.
    :param str risk: the channel risk get the snap from.
    :param str track: the specific channel track get the snap from.
    :param str download_path: the path to write the downloaded snap to.
    :param str arch: the architecture of the download as a deb arch.
    :param str except_hash: do not download if set to a sha3_384 hash that
                            matches the snap_name to be downloaded.
    :raises storeapi.errors.SHAMismatchErrorRuntimeError:
         If the checksum for the downloaded file does not match the expected
         hash.
    :returns: A sha3_384 of the file that was or would have been downloaded.
    """
    store = storeapi.StoreClient()
    return store.download(
        snap_name,
        risk=risk,
        track=track,
        download_path=download_path,
        arch=arch,
        except_hash=except_hash,
    )


def status(snap_name, series, arch):
    store = storeapi.StoreClient()

    with _requires_login():
        status = store.get_snap_status(snap_name, series, arch)

    channel_map_tree = status.get("channel_map_tree", {})
    # This does not look good in green so we print instead
    tabulated_status = _tabulated_channel_map_tree(channel_map_tree)
    print(tabulated_status)


def _get_text_for_current_channels(channels, current_channels):
    return (
        ", ".join(
            channel + ("*" if channel in current_channels else "")
            for channel in channels
        )
        or "-"
    )


def revisions(snap_name, series, arch):
    store = storeapi.StoreClient()

    with _requires_login():
        revisions = store.get_snap_revisions(snap_name, series, arch)

    parsed_revisions = [
        (
            rev["revision"],
            rev["timestamp"],
            rev["arch"],
            rev["version"],
            _get_text_for_current_channels(rev["channels"], rev["current_channels"]),
        )
        for rev in revisions
    ]
    tabulated_revisions = tabulate(
        parsed_revisions,
        numalign="left",
        headers=["Rev.", "Uploaded", "Arch", "Version", "Channels"],
        tablefmt="plain",
    )
    print(tabulated_revisions)


def gated(snap_name):
    """Print list of snaps gated by snap_name."""
    store = storeapi.StoreClient()
    # Get data for the gating snap
    with _requires_login():
        snaps = store.get_account_information().get("snaps", {})

    release = storeapi.constants.DEFAULT_SERIES
    # Resolve name to snap-id
    try:
        snap_id = snaps[release][snap_name]["snap-id"]
    except KeyError:
        raise storeapi.errors.SnapNotFoundError(snap_name)

    validations = store.get_assertion(snap_id, endpoint="validations")

    if validations:
        table_data = []
        for v in validations:
            name = v["approved-snap-name"]
            revision = v["approved-snap-revision"]
            if revision == "-":
                revision = None
            required = str(v.get("required", True))
            # Currently timestamps have microseconds, which look bad
            timestamp = v["timestamp"]
            if "." in timestamp:
                timestamp = timestamp.split(".")[0] + "Z"
            table_data.append([name, revision, required, timestamp])
        tabulated = tabulate(
            table_data,
            headers=["Name", "Revision", "Required", "Approved"],
            tablefmt="plain",
            missingval="-",
        )
        print(tabulated)
    else:
        print("There are no validations for snap {!r}".format(snap_name))


def validate(snap_name, validations, revoke=False, key=None):
    """Generate, sign and upload validation assertions."""
    # Check validations format
    _check_validations(validations)

    store = storeapi.StoreClient()

    # Need the ID of the logged in user.
    with _requires_login():
        account_info = store.get_account_information()
    authority_id = account_info["account_id"]

    # Get data for the gating snap
    release = storeapi.constants.DEFAULT_SERIES
    try:
        snap_id = account_info["snaps"][release][snap_name]["snap-id"]
    except KeyError:
        raise storeapi.errors.SnapNotFoundError(snap_name)

    # Then, for each requested validation, generate assertion
    for validation in validations:
        gated_name, rev = validation.split("=", 1)
        echo.info("Getting details for {}".format(gated_name))
        approved_data = store.cpi.get_info(gated_name)
        assertion = {
            "type": "validation",
            "authority-id": authority_id,
            "series": release,
            "snap-id": snap_id,
            "approved-snap-id": approved_data.snap_id,
            "approved-snap-revision": rev,
            "timestamp": datetime.utcnow().isoformat() + "Z",
            "revoked": "false",
        }
        if revoke:
            assertion["revoked"] = "true"

        assertion = _sign_assertion(validation, assertion, key, "validations")

        # Save assertion to a properly named file
        fname = "{}-{}-r{}.assertion".format(snap_name, gated_name, rev)
        with open(fname, "wb") as f:
            f.write(assertion)

        store.push_assertion(snap_id, assertion, endpoint="validations")


validation_re = re.compile("^[^=]+=[0-9]+$")


def _check_validations(validations):
    invalids = [v for v in validations if not validation_re.match(v)]
    if invalids:
        raise storeapi.errors.InvalidValidationRequestsError(invalids)


def _sign_assertion(snap_name, assertion, key, endpoint):
    cmdline = ["snap", "sign"]
    if key:
        cmdline += ["-k", key]
    snap_sign = Popen(
        cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    data = json.dumps(assertion).encode("utf8")
    echo.info("Signing {} assertion for {}".format(endpoint, snap_name))
    assertion, err = snap_sign.communicate(input=data)
    if snap_sign.returncode != 0:
        err = err.decode()
        raise storeapi.errors.StoreAssertionError(
            endpoint=endpoint, snap_name=snap_name, error=err
        )

    return assertion



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import subprocess

from ._deltas import BaseDeltasGenerator

logger = logging.getLogger(__name__)


_DELTA_TOOL = "xdelta3"
_DELTA_FORMAT = "xdelta3"
_DELTA_EXTNAME = "xdelta3"


class XDelta3Generator(BaseDeltasGenerator):
    def __init__(self, *, source_path, target_path):
        super().__init__(
            source_path=source_path,
            target_path=target_path,
            delta_tool=_DELTA_TOOL,
            delta_format=_DELTA_FORMAT,
            delta_file_extname=_DELTA_EXTNAME,
        )

    def get_delta_cmd(self, source_path, target_path, delta_file):
        return [self.delta_tool_path, "-s", source_path, target_path, delta_file]

    def log_delta_file(self, delta_file):
        xdelta_output = subprocess.check_output(
            [self.delta_tool_path, "printhdr", delta_file], universal_newlines=True
        )
        logger.debug("xdelta3 delta diff generation:\n{}".format(xdelta_output))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os

from snapcraft.internal import common, errors
from snapcraft.formatting_utils import humanize_list


class BuilderEnvironmentConfig:
    """Handle the chosen build provider.

    To determine the build environment, SNAPCRAFT_BUILD_ENVIRONMENT is
    retrieved from the environment and used to determine the build
    provider. An environment can be forced by use of force_provider
    during initialization. The default environment is multipass
    unless the snapcraft is run inside docker for which the preferred
    environment will be to use the host.

    Valid values are:

    - host: the host will drive the build.
    - managed-host: the host will drive the build, but work best for discardable
                    build providers.
    - multipass: a vm driven by multipass will be created to drive the build.
    """

    def __init__(self, *, force_provider: str = None) -> None:
        """Instantiate a BuildEnvironmentConfig.

        :param str force_provider: ignore the hints from the environment and use
                                   the specified provider.
        """
        if force_provider:
            build_provider = force_provider
        elif common.is_process_container():
            build_provider = "host"
        else:
            build_provider = os.environ.get("SNAPCRAFT_BUILD_ENVIRONMENT", "multipass")

        valid_providers = ["host", "multipass", "managed-host", "lxd"]
        if build_provider not in valid_providers:
            raise errors.SnapcraftEnvironmentError(
                "The snapcraft build environment must be one of: {}.".format(
                    humanize_list(items=valid_providers, conjunction="or")
                )
            )

        self.provider = build_provider
        self.is_host = build_provider == "host"
        self.is_multipass = build_provider == "multipass"
        self.is_multipass = build_provider == "lxd"
        self.is_managed_host = build_provider == "managed-host"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import subprocess
import sys
from typing import List, Sequence

import snapcraft
from snapcraft.internal import errors, repo

logger = logging.getLogger(__name__)


class WstoolError(errors.SnapcraftError):
    pass


class WorkspaceInitializationError(WstoolError):
    fmt = "Error initializing workspace: {message}"

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class RosinstallMergeError(WstoolError):
    fmt = "Error merging rosinstall file {path!r} into workspace: {message}"

    def __init__(self, path: str, message: str) -> None:
        super().__init__(path=path, message=message)


class WorkspaceUpdateError(WstoolError):
    fmt = "Error updating workspace: {message}"

    def __init__(self, message: str) -> None:
        super().__init__(message=message)


class Wstool:
    """This class serves as a Python wrapper for the CLI utility wstool."""

    def __init__(
        self,
        ros_package_path: str,
        wstool_path: str,
        ubuntu_sources: str,
        ubuntu_keyrings: Sequence[str],
        project: snapcraft.ProjectOptions,
    ) -> None:
        """Create new Wstool

        :param str ros_package_path: The path where the packages should be
                                     fetched.
        :param str wstool_path: Working directory for wstool (where it will be
                                installed).
        :param str ubuntu_sources: Ubuntu repositories from which wstool will
                                   be installed.
        :param project: Instance of ProjectOptions for project-wide settings.
        :type project: snapcraft.ProjectOptions
        """
        self._ros_package_path = ros_package_path
        self._ubuntu_sources = ubuntu_sources
        self._ubuntu_keyrings = ubuntu_keyrings
        self._wstool_path = wstool_path
        self._wstool_install_path = os.path.join(wstool_path, "install")
        self._project = project

    def setup(self) -> None:
        """Download/install wstool, and initialize workspace."""

        os.makedirs(self._wstool_install_path, exist_ok=True)

        # wstool isn't a dependency of the project, so we'll unpack it
        # somewhere else, and use it from there.
        logger.info("Preparing to fetch wstool...")
        ubuntu = repo.Ubuntu(
            self._wstool_path,
            sources=self._ubuntu_sources,
            keyrings=self._ubuntu_keyrings,
            project_options=self._project,
        )
        logger.info("Fetching wstool...")
        ubuntu.get(["python-wstool"])

        logger.info("Installing wstool...")
        ubuntu.unpack(self._wstool_install_path)

        logger.info("Initializing workspace (if necessary)...")
        try:
            self._run(
                [
                    "init",
                    self._ros_package_path,
                    "-j{}".format(self._project.parallel_build_count),
                ]
            )
        except subprocess.CalledProcessError as e:
            output = e.output.decode(sys.getfilesystemencoding()).strip()
            if "already is a workspace" not in output:
                stderr = e.stderr.decode(sys.getfilesystemencoding()).strip()
                raise WorkspaceInitializationError(stderr)

    def merge(self, rosinstall_file: str) -> str:
        """Merge rosinstall file with existing rosinstall in workspace.

        :param str rosinstall_file: Path to rosinstall file to merge in.
        """

        try:
            # Summary of arguments used:
            # --confirm-all: Don't ask for confirmation
            # -t: Path to workspace
            return self._run(
                [
                    "merge",
                    rosinstall_file,
                    "--confirm-all",
                    "-t{}".format(self._ros_package_path),
                ]
            ).strip()
        except subprocess.CalledProcessError as e:
            stderr = e.stderr.decode(sys.getfilesystemencoding()).strip()
            raise RosinstallMergeError(rosinstall_file, stderr)

    def update(self) -> str:
        """Update workspace from rosinstall file.

        This actually hits the network and downloads all repositories in
        the workspace's rosinstall file.
        """
        try:
            # Summary of arguments used:
            # -j: Use multiple jobs to fetch
            # -t: Path to workspace (into which packages will be fetched)
            return self._run(
                [
                    "update",
                    "-j{}".format(self._project.parallel_build_count),
                    "-t{}".format(self._ros_package_path),
                ]
            )
        except subprocess.CalledProcessError as e:
            stderr = e.stderr.decode(sys.getfilesystemencoding()).strip()
            raise WorkspaceUpdateError(stderr)

    def _run(self, arguments: List[str]) -> str:
        env = os.environ.copy()

        env["PATH"] += ":" + os.path.join(self._wstool_install_path, "usr", "bin")
        env["PYTHONPATH"] = os.path.join(
            self._wstool_install_path, "usr", "lib", "python2.7", "dist-packages"
        )

        return (
            subprocess.check_output(
                ["wstool"] + arguments, stderr=subprocess.PIPE, env=env
            )
            .decode(sys.getfilesystemencoding())
            .strip()
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from snapcraft import yaml_utils
from snapcraft.internal import steps


class State(yaml_utils.SnapcraftYAMLObject):
    def __repr__(self):
        items = sorted(self.__dict__.items())
        strings = (": ".join((key, repr(value))) for key, value in items)
        representation = ", ".join(strings)

        return "{}({})".format(self.__class__.__name__, representation)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__

        return False


class PartState(State):
    def __init__(self, part_properties, project):
        super().__init__()
        if not part_properties:
            part_properties = {}

        self.properties = self.properties_of_interest(part_properties)
        self.project_options = self.project_options_of_interest(project)

    def properties_of_interest(self, part_properties):
        """Extract the properties concerning this step from the options.

        Note that these options come from the YAML for a given part.
        """

        raise NotImplementedError

    def project_options_of_interest(self, project):
        """Extract the options concerning this step from the project."""

        raise NotImplementedError

    def diff_properties_of_interest(self, other_properties):
        """Return set of properties that differ."""

        return _get_differing_keys(
            self.properties, self.properties_of_interest(other_properties)
        )

    def diff_project_options_of_interest(self, other_project_options):
        """Return set of project options that differ."""

        return _get_differing_keys(
            self.project_options,
            self.project_options_of_interest(other_project_options),
        )


def _get_differing_keys(dict1, dict2):
    differing_keys = set()
    for key, dict1_value in dict1.items():
        dict2_value = dict2.get(key)
        if dict1_value != dict2_value:
            differing_keys.add(key)

    for key, dict2_value in dict2.items():
        dict1_value = dict1.get(key)
        if dict1_value != dict2_value:
            differing_keys.add(key)

    return differing_keys


def get_state(state_dir: str, step: steps.Step):
    state = None
    state_file = get_step_state_file(state_dir, step)
    if os.path.isfile(state_file):
        with open(state_file, "r") as f:
            state = yaml_utils.load(f)

    return state


def get_step_state_file(state_dir: str, step: steps.Step) -> str:
    return os.path.join(state_dir, step.name)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import subprocess
import tempfile
from datetime import datetime
from textwrap import dedent
from typing import List, Dict

import click

import snapcraft
from . import echo
from snapcraft import storeapi, yaml_utils
from snapcraft.storeapi import assertions


_COLLABORATION_HEADER = dedent(
    """\
    # Change which developers may push or release snaps on the publisher's behalf.
    #
    # Sample entry:
    #
    # developers:
    #   - developer-id: "dev-one"      # Which developer
    #     since: "2017-02-10 08:35:00" # When contributions started
    #     until: "2018-02-10 08:35:00" # When contributions ceased (optional)
    #
    # All timestamps are UTC, and the "now" special string will be replaced by
    # the current time. Do not remove entries or use an until time in the past
    # unless you want existing snaps provided by the developer to stop working."""
)  # noqa


@click.group()
def assertionscli():
    """Store assertion commands"""
    pass


@assertionscli.command("list-keys")
def list_keys():
    """List the keys available to sign assertions.

    This command has an alias of `keys`.
    """
    snapcraft.list_keys()


@assertionscli.command("create-key")
@click.argument("key-name", metavar="<key-name>", required=False)
def create_key(key_name: str) -> None:
    """Create a key to sign assertions."""
    snapcraft.create_key(key_name)


@assertionscli.command("register-key")
@click.argument("key-name", metavar="<key-name>", required=False)
def register_key(key_name: str) -> None:
    """Register a key with the store to sign assertions."""
    snapcraft.register_key(key_name)


@assertionscli.command("sign-build")
@click.option("--key-name", metavar="<key-name>")
@click.argument(
    "snap-file",
    metavar="<snap-file>",
    type=click.Path(exists=True, readable=True, resolve_path=True, dir_okay=False),
)
@click.option(
    "--local", is_flag=True, help="Do not push the generated assertion to the store"
)
def sign_build(snap_file: str, key_name: str, local: bool) -> None:
    """Sign a built snap file and assert it using the developer's key."""
    snapcraft.sign_build(snap_file, key_name=key_name, local=local)


@assertionscli.command()
@click.argument("snap-name", metavar="<snap-name>")
@click.argument("validations", metavar="<validation>...", nargs=-1, required=True)
@click.option("--key-name", metavar="<key-name>")
def validate(snap_name: str, validations: list, key_name: str) -> None:
    """Validate a gated snap."""
    snapcraft.validate(snap_name, validations, key=key_name)


@assertionscli.command()
@click.argument("snap-name", metavar="<snap-name>")
def gated(snap_name: str) -> None:
    """Get the list of snaps and revisions gating a snap."""
    snapcraft.gated(snap_name)


@assertionscli.command("edit-collaborators")
@click.argument("snap-name", metavar="<snap-name>")
@click.option("--key-name", metavar="<key-name>")
def edit_collaborators(snap_name, key_name):
    """Edit the list of collaborators for <snap-name>.

    This command has an alias of `collaborators`.
    """
    dev_assertion = assertions.DeveloperAssertion(
        snap_name=snap_name, signing_key=key_name
    )
    developers = dev_assertion.get_developers()
    updated_developers = _update_developers(developers)
    new_dev_assertion = dev_assertion.new_assertion(developers=updated_developers)
    if new_dev_assertion.is_equal(dev_assertion):
        echo.warning("Aborting due to unchanged collaborators list.")
        return

    try:
        new_dev_assertion.push()
    except storeapi.errors.StoreValidationError as store_error:
        if store_error.error_list[0]["code"] != "revoked-uploads":
            raise store_error
        click.echo(
            "This will revoke the following collaborators: {!r}".format(
                " ".join(store_error.error_list[0]["extra"])
            )
        )
        if echo.confirm("Are you sure you want to continue?"):
            new_dev_assertion.push(force=True)
        else:
            echo.warning("The collaborators for this snap have not been altered.")


def _update_developers(developers: List[Dict[str, str]]) -> List[Dict[str, str]]:
    edit_friendly_developers = _reformat_time_for_editing(developers)
    updated_developers = _edit_developers(edit_friendly_developers)
    return _reformat_time_for_assertion(updated_developers)


def _edit_developers(developers: List[Dict[str, str]]) -> List[Dict[str, str]]:
    """Spawn an editor to modify the snap-developer assertion for a snap."""
    editor_cmd = os.getenv("EDITOR", "vi")

    developer_wrapper = {"developers": developers}

    with tempfile.NamedTemporaryFile() as ft:
        ft.close()
        with open(ft.name, "w") as fw:
            print(_COLLABORATION_HEADER, file=fw)
            yaml_utils.dump(developer_wrapper, stream=fw)
        subprocess.check_call([editor_cmd, ft.name])
        with open(ft.name, "r") as fr:
            developers = yaml_utils.load(fr).get("developers")
    return developers


def _reformat_time_for_editing(
    developers: List[Dict[str, str]], time_format: str = "%Y-%m-%d %H:%M:%S"
) -> List[Dict[str, str]]:
    reformatted_developers = []
    for developer in developers:
        developer_it = {"developer-id": developer["developer-id"]}
        for range in ["since", "until"]:
            if range in developer:
                date = datetime.strptime(developer[range], "%Y-%m-%dT%H:%M:%S.%fZ")
                developer_it[range] = datetime.strftime(date, time_format)
        reformatted_developers.append(developer_it)
    return reformatted_developers


def _reformat_time_for_assertion(
    developers: List[Dict[str, str]]
) -> List[Dict[str, str]]:
    reformatted_developers = []
    for developer in developers:
        developer_it = {"developer-id": developer["developer-id"]}
        for range_ in ["since", "until"]:
            if range_ in developer:
                if developer[range_] == "now":
                    date = datetime.now()
                else:
                    date = datetime.strptime(developer[range_], "%Y-%m-%d %H:%M:%S")
                # We don't care about microseconds because we cannot edit
                # later so we set that to 0.
                developer_it[range_] = date.strftime("%Y-%m-%dT%H:%M:%S.000000Z")
        reformatted_developers.append(developer_it)
    return reformatted_developers



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import contextlib
import copy
import filecmp
import logging
import os
import shutil
import subprocess
import sys
from glob import glob, iglob
from typing import cast, Dict, List, Set, Sequence

import snapcraft.extractors
from snapcraft import file_utils, yaml_utils
from snapcraft.internal import common, elf, errors, repo, sources, states, steps
from snapcraft.internal.mangling import clear_execstack

from ._build_attributes import BuildAttributes
from ._dependencies import MissingDependencyResolver
from ._metadata_extraction import extract_metadata
from ._plugin_loader import load_plugin  # noqa
from ._runner import Runner
from ._patchelf import PartPatcher
from ._dirty_report import Dependency, DirtyReport  # noqa
from ._outdated_report import OutdatedReport

logger = logging.getLogger(__name__)


class PluginHandler:
    @property
    def name(self):
        return self.plugin.name

    def __init__(
        self,
        *,
        plugin,
        part_properties,
        project_options,
        part_schema,
        definitions_schema,
        stage_packages_repo,
        grammar_processor,
        snap_base_path,
        base,
        confinement,
        snap_type,
        soname_cache
    ):
        self.valid = False
        self.plugin = plugin
        self._part_properties = _expand_part_properties(part_properties, part_schema)
        self.stage_packages = []
        self._stage_packages_repo = stage_packages_repo
        self._grammar_processor = grammar_processor
        self._snap_base_path = snap_base_path
        self._base = base
        self._confinement = confinement
        self._snap_type = snap_type
        self._soname_cache = soname_cache
        self._source = grammar_processor.get_source()
        if not self._source:
            self._source = part_schema["source"].get("default")

        self._pull_state = None  # type: states.PullState
        self._build_state = None  # type: states.BuildState
        self._stage_state = None  # type: states.StageState
        self._prime_state = None  # type: states.PrimeState

        self._project_options = project_options
        self.deps = []

        self.stagedir = project_options.stage_dir
        self.primedir = project_options.prime_dir

        # We don't need to set the source_handler on systems where we do not
        # build
        if sys.platform == "linux":
            self.source_handler = self._get_source_handler(self._part_properties)
        else:
            self.source_handler = None

        self.build_environment = _list_of_dicts_to_env(
            self._part_properties["build-environment"]
        )

        self._build_attributes = BuildAttributes(
            self._part_properties["build-attributes"]
        )

        # Scriptlet data is a dict of dicts for each step
        self._scriptlet_metadata = collections.defaultdict(
            snapcraft.extractors.ExtractedMetadata
        )
        self._runner = Runner(
            part_properties=self._part_properties,
            sourcedir=self.plugin.sourcedir,
            builddir=self.plugin.build_basedir,
            stagedir=self.stagedir,
            primedir=self.primedir,
            builtin_functions={
                steps.PULL.name: self._do_pull,
                steps.BUILD.name: self.plugin.build,
                steps.STAGE.name: self._do_stage,
                steps.PRIME.name: self._do_prime,
                "set-version": self._set_version,
                "set-grade": self._set_grade,
            },
        )

        self._migrate_state_file()
        self._current_step = None  # type: steps.Step

    def get_pull_state(self) -> states.PullState:
        if not self._pull_state:
            self._pull_state = cast(states.PullState, self.get_state(steps.PULL))
        return self._pull_state

    def get_build_state(self) -> states.BuildState:
        if not self._build_state:
            self._build_state = cast(states.BuildState, self.get_state(steps.BUILD))
        return self._build_state

    def get_stage_state(self) -> states.StageState:
        if not self._stage_state:
            self._stage_state = cast(states.StageState, self.get_state(steps.STAGE))
        return self._stage_state

    def get_prime_state(self) -> states.PrimeState:
        if not self._prime_state:
            self._prime_state = cast(states.PrimeState, self.get_state(steps.PRIME))
        return self._prime_state

    def get_state(self, step) -> states.PartState:
        return states.get_state(self.plugin.statedir, step)

    def _get_source_handler(self, properties):
        """Returns a source_handler for the source in properties."""
        # TODO: we cannot pop source as it is used by plugins. We also make
        # the default '.'
        source_handler = None
        if self._source:
            handler_class = sources.get_source_handler(
                self._source, source_type=properties["source-type"]
            )
            source_handler = handler_class(
                self._source,
                self.plugin.sourcedir,
                source_checksum=properties["source-checksum"],
                source_branch=properties["source-branch"],
                source_tag=properties["source-tag"],
                source_depth=properties["source-depth"],
                source_commit=properties["source-commit"],
            )

        return source_handler

    def _set_version(self, *, version):
        try:
            self._set_scriptlet_metadata(
                snapcraft.extractors.ExtractedMetadata(version=version)
            )
        except errors.ScriptletDuplicateDataError as e:
            raise errors.ScriptletDuplicateFieldError("version", e.other_step)

    def _set_grade(self, *, grade):
        try:
            self._set_scriptlet_metadata(
                snapcraft.extractors.ExtractedMetadata(grade=grade)
            )
        except errors.ScriptletDuplicateDataError as e:
            raise errors.ScriptletDuplicateFieldError("grade", e.other_step)

    def _check_scriplet_metadata_dupe(
        self, metadata: snapcraft.extractors.ExtractedMetadata, step: steps.Step
    ):
        # First, ensure the metadata set here doesn't conflict with metadata
        # already set for this step
        conflicts = metadata.overlap(self._scriptlet_metadata[step])
        if len(conflicts) > 0:
            raise errors.ScriptletDuplicateDataError(step, step, list(conflicts))

        # Now ensure the metadata from this step doesn't conflict with
        # metadata from any other step (if any)
        with contextlib.suppress(errors.NoLatestStepError):
            latest_step = self.latest_step()
            required_steps = latest_step.previous_steps() + [latest_step]
            for other_step in reversed(required_steps):
                state = states.get_state(self.plugin.statedir, other_step)
                conflicts = metadata.overlap(state.scriptlet_metadata)
                if len(conflicts) > 0:
                    raise errors.ScriptletDuplicateDataError(
                        step, other_step, list(conflicts)
                    )

    def _set_scriptlet_metadata(self, metadata: snapcraft.extractors.ExtractedMetadata):
        try:
            step = self.next_step()
            self._check_scriplet_metadata_dupe(metadata, step)
        except errors.NoNextStepError:
            # We've already run through all the steps, don't bother to
            # check for duplication as it is not really necessary and
            # _check_scriplet_metadata_dupe() assumes that all steps
            # haven't already executed.  Determine the current step
            # as saved by _do_runner_step().
            step = self._current_step

        self._scriptlet_metadata[step].update(metadata)

    def makedirs(self):
        dirs = [
            self.plugin.sourcedir,
            self.plugin.builddir,
            self.plugin.installdir,
            self.plugin.statedir,
            self.stagedir,
            self.primedir,
        ]
        for d in dirs:
            os.makedirs(d, exist_ok=True)

    def _do_runner_step(self, step: steps.Step):
        self.makedirs()
        self._current_step = step
        return getattr(self._runner, "{}".format(step.name))()

    def _migrate_state_file(self):
        # In previous versions of Snapcraft, the state directory was a file.
        # Rather than die if we're running on output from an old version,
        # migrate it for them.
        if os.path.isfile(self.plugin.statedir):
            with open(self.plugin.statedir, "r") as f:
                step = f.read()

            if step:
                os.remove(self.plugin.statedir)
                os.makedirs(self.plugin.statedir)
                self.mark_done(steps.get_step_by_name(step))

    def working_directory_for_step(self, step: steps.Step) -> str:
        if step == steps.PULL:
            return self.plugin.sourcedir
        elif step == steps.BUILD:
            return self.plugin.builddir
        elif step == steps.STAGE:
            return self.stagedir
        elif step == steps.PRIME:
            return self.primedir

        raise errors.InvalidStepError(step.name)

    def latest_step(self):
        for step in reversed(steps.STEPS):
            if os.path.exists(states.get_step_state_file(self.plugin.statedir, step)):
                return step

        raise errors.NoLatestStepError(self.name)

    def next_step(self):
        latest_step = None
        with contextlib.suppress(errors.NoLatestStepError):
            latest_step = self.latest_step()
        next_step = steps.next_step(latest_step)
        if not next_step:
            raise errors.NoNextStepError(self.name)
        return next_step

    def is_clean(self, step):
        """Return true if the given step hasn't run (or has been cleaned)."""

        try:
            latest_step = self.latest_step()
            return step > latest_step
        except errors.NoLatestStepError:
            return True

    def is_outdated(self, step: steps.Step) -> bool:
        """Return true if the given step is outdated.

        See get_outdated_report() for more information.

        :param steps.Step step: The step to be checked.
        """

        return self.get_outdated_report(step) is not None

    def get_outdated_report(self, step: steps.Step) -> OutdatedReport:
        """Return an OutdatedReport class describing why the step is outdated.

        A step is considered to be outdated if an earlier step in the lifecycle
        has been run more recently, or if the source code changed on disk.
        This means the step needs to be updated by taking modified files from
        the previous step. This is in contrast to a "dirty" step, which must
        be cleaned and run again.

        :param steps.Step step: The step to be checked.
        :returns: OutdatedReport if the step is outdated, None otherwise.
        """

        try:
            return getattr(self, "check_{}".format(step.name))()
        except AttributeError:
            with contextlib.suppress(errors.StepHasNotRunError):
                timestamp = self.step_timestamp(step)

                for previous_step in reversed(step.previous_steps()):
                    # Has a previous step run since this one ran? Then this
                    # step needs to be updated.
                    with contextlib.suppress(errors.StepHasNotRunError):
                        if timestamp < self.step_timestamp(previous_step):
                            return OutdatedReport(previous_step_modified=previous_step)
            return None

    def is_dirty(self, step: steps.Step) -> bool:
        """Return true if the given step is dirty.

        See get_dirty_report() for more information.

        :param steps.Step step: The step to be checked.
        """

        return self.get_dirty_report(step) is not None

    def get_dirty_report(self, step: steps.Step) -> DirtyReport:
        """Return a DirtyReport class describing why the step is dirty.

        A step is considered to be dirty if either YAML properties used by it
        (`stage-packages` are used by the `pull` step, for example), or project
        options used by it (`--target-arch` is used by the `pull` step as well)
        have changed since the step was run. This means the step needs to be
        cleaned and run again. This is in contrast to an "outdated" step, which
        typically doesn't need to be cleaned, just updated with files from an
        earlier step in the lifecycle.

        :param steps.Step step: The step to be checked.
        :returns: DirtyReport if the step is dirty, None otherwise.
        """

        # Retrieve the stored state for this step (assuming it has already run)
        state = states.get_state(self.plugin.statedir, step)
        if state:
            # state.properties contains the old YAML that this step cares
            # about, and we're comparing it to those same keys in the current
            # YAML (self._part_properties). If they've changed, then this step
            # is dirty and needs to run again.
            properties = state.diff_properties_of_interest(self._part_properties)

            # state.project_options contains the old project options that this
            # step cares about, and we're comparing it to those same options in
            # the current project. If they've changed, then this step is dirty
            # and needs to run again.
            options = state.diff_project_options_of_interest(self._project_options)

            if properties or options:
                return DirtyReport(
                    dirty_properties=properties, dirty_project_options=options
                )

        return None

    def should_step_run(self, step, force=False):
        return force or self.is_clean(step)

    def step_timestamp(self, step):
        try:
            return os.stat(
                states.get_step_state_file(self.plugin.statedir, step)
            ).st_mtime
        except FileNotFoundError as e:
            raise errors.StepHasNotRunError(self.name, step) from e

    def mark_done(self, step, state=None):
        if not state:
            state = {}

        with open(states.get_step_state_file(self.plugin.statedir, step), "w") as f:
            f.write(yaml_utils.dump(state))

    def mark_cleaned(self, step):
        state_file = states.get_step_state_file(self.plugin.statedir, step)
        if os.path.exists(state_file):
            os.remove(state_file)

        if os.path.isdir(self.plugin.statedir) and not os.listdir(self.plugin.statedir):
            os.rmdir(self.plugin.statedir)

    def _fetch_stage_snaps(self):
        stage_snaps = self._grammar_processor.get_stage_snaps()
        if stage_snaps:
            repo.snaps.download_snaps(
                snaps_list=stage_snaps, directory=self.plugin.snapsdir
            )

    def _unpack_stage_snaps(self):
        stage_snaps = self._grammar_processor.get_stage_snaps()
        if not stage_snaps:
            return

        logger.debug("Unpacking stage-snaps to {!r}".format(self.plugin.stage_snaps))
        snap_files = iglob(os.path.join(self.plugin.snapsdir, "*.snap"))
        snap_sources = (
            sources.Snap(source=s, source_dir=self.plugin.snapsdir) for s in snap_files
        )
        for snap_source in snap_sources:
            snap_source.provision(
                self.plugin.installdir, clean_target=False, keep_snap=True
            )

    def _fetch_stage_packages(self):
        stage_packages = self._grammar_processor.get_stage_packages()
        if stage_packages:
            logger.debug("Fetching stage-packages {!r}".format(stage_packages))
            try:
                self.stage_packages = self._stage_packages_repo.get(stage_packages)
            except repo.errors.PackageNotFoundError as e:
                raise errors.StagePackageDownloadError(self.name, e.message)

    def _unpack_stage_packages(self):
        stage_packages = self._grammar_processor.get_stage_packages()
        if stage_packages:
            logger.debug(
                "Unpacking stage-packages to {!r}".format(self.plugin.installdir)
            )
            self._stage_packages_repo.unpack(self.plugin.installdir)

    def prepare_pull(self, force=False):
        self.makedirs()
        self._fetch_stage_packages()
        self._fetch_stage_snaps()
        self._unpack_stage_packages()
        self._unpack_stage_snaps()

    def pull(self, force=False):
        # Ensure any previously-failed pull is cleared out before we try again
        if os.path.islink(self.plugin.sourcedir) or os.path.isfile(
            self.plugin.sourcedir
        ):
            os.remove(self.plugin.sourcedir)
        elif os.path.isdir(self.plugin.sourcedir):
            shutil.rmtree(self.plugin.sourcedir)

        self._do_runner_step(steps.PULL)
        self.mark_pull_done()

    def check_pull(self):
        # Check to see if pull needs to be updated
        state_file = states.get_step_state_file(self.plugin.statedir, steps.PULL)

        # Not all sources support checking for updates
        with contextlib.suppress(sources.errors.SourceUpdateUnsupportedError):
            if self.source_handler.check(state_file):
                return OutdatedReport(source_updated=True)
        return None

    def update_pull(self):
        self._do_runner_step(steps.PULL)
        self.source_handler.update()
        self.mark_pull_done()

    def _do_pull(self):
        if self.source_handler:
            self.source_handler.pull()
        self.plugin.pull()

    def mark_pull_done(self):
        pull_properties = self.plugin.get_pull_properties()

        # Add the annotated list of build packages
        part_build_packages = self._part_properties.get("build-packages", [])
        part_build_snaps = self._part_properties.get("build-snaps", [])

        # Extract any requested metadata available in the source directory
        metadata = snapcraft.extractors.ExtractedMetadata()
        metadata_files = []
        for parse_relpath in self._part_properties.get("parse-info", []):
            with contextlib.suppress(errors.MissingMetadataFileError):
                metadata.update(
                    extract_metadata(self.name, parse_relpath, self.plugin.sourcedir)
                )
                metadata_files.append(
                    os.path.join(self.plugin.sourcedir, parse_relpath)
                )

        self.mark_done(
            steps.PULL,
            states.PullState(
                pull_properties,
                part_properties=self._part_properties,
                project=self._project_options,
                stage_packages=self.stage_packages,
                build_snaps=part_build_snaps,
                build_packages=part_build_packages,
                source_details=getattr(self.source_handler, "source_details", None),
                metadata=metadata,
                metadata_files=metadata_files,
                scriptlet_metadata=self._scriptlet_metadata[steps.PULL],
            ),
        )

    def clean_pull(self):
        if self.is_clean(steps.PULL):
            return

        # Remove ubuntu cache (where stage packages are fetched)
        if os.path.exists(self.plugin.osrepodir):
            shutil.rmtree(self.plugin.osrepodir)

        # Remove snaps dir (where stage snaps are fetched)
        if os.path.exists(self.plugin.snapsdir):
            shutil.rmtree(self.plugin.snapsdir)

        if os.path.exists(self.plugin.sourcedir):
            if os.path.islink(self.plugin.sourcedir):
                os.remove(self.plugin.sourcedir)
            else:
                shutil.rmtree(self.plugin.sourcedir)

        self.plugin.clean_pull()
        self.mark_cleaned(steps.PULL)

    def prepare_build(self, force=False):
        self.makedirs()
        # Stage packages are fetched and unpacked in the pull step, but we'll
        # unpack again here just in case the build step has been cleaned.
        self._unpack_stage_packages()

    def build(self, force=False):
        self.makedirs()

        if not self.plugin.out_of_source_build:
            if os.path.exists(self.plugin.build_basedir):
                shutil.rmtree(self.plugin.build_basedir)

            # FIXME: It's not necessary to ignore here anymore since it's now
            # done in the Local source. However, it's left here so that it
            # continues to work on old snapcraft trees that still have src
            # symlinks.
            def ignore(directory, files):
                if directory == self.plugin.sourcedir:
                    snaps = glob(os.path.join(directory, "*.snap"))
                    if snaps:
                        snaps = [os.path.basename(s) for s in snaps]
                        return common.SNAPCRAFT_FILES + snaps
                    else:
                        return common.SNAPCRAFT_FILES
                else:
                    return []

            # No hard-links being used here in case the build process modifies
            # these files.
            shutil.copytree(
                self.plugin.sourcedir,
                self.plugin.build_basedir,
                symlinks=True,
                ignore=ignore,
            )

        self._do_build()

    def update_build(self):
        if not self.plugin.out_of_source_build:
            # Use the local source to update. It's important to use
            # file_utils.copy instead of link_or_copy, as the build process
            # may modify these files
            source = sources.Local(
                self.plugin.sourcedir,
                self.plugin.build_basedir,
                copy_function=file_utils.copy,
            )
            if not source.check(
                states.get_step_state_file(self.plugin.statedir, steps.BUILD)
            ):
                return
            source.update()

        self._do_build(update=True)

    def _do_build(self, *, update=False):
        self._do_runner_step(steps.BUILD)

        # Organize the installed files as requested. We do this in the build step for
        # two reasons:
        #
        #   1. So cleaning and re-running the stage step works even if `organize` is
        #      used
        #   2. So collision detection takes organization into account, i.e. we can use
        #      organization to get around file collisions between parts when staging.
        #
        # If `update` is true, we give the snapcraft CLI permission to overwrite files
        # that already exist. Typically we do NOT want this, so that parts don't
        # accidentally clobber e.g. files brought in from stage-packages, but in the
        # case of updating build, we want the part to have the ability to organize over
        # the files it organized last time around. We can be confident that this won't
        # overwrite anything else, because to do so would require changing the
        # `organize` keyword, which will make the build step dirty and require a clean
        # instead of an update.
        self._organize(overwrite=update)

        self.mark_build_done()

    def mark_build_done(self):
        build_properties = self.plugin.get_build_properties()
        plugin_manifest = self.plugin.get_manifest()
        machine_manifest = self._get_machine_manifest()

        # Extract any requested metadata available in the build directory,
        # followed by the install directory (which takes precedence)
        metadata_files = []
        metadata = snapcraft.extractors.ExtractedMetadata()
        for parse_relpath in self._part_properties.get("parse-info", []):
            found_path = None
            with contextlib.suppress(errors.MissingMetadataFileError):
                metadata.update(
                    extract_metadata(self.name, parse_relpath, self.plugin.builddir)
                )
                found_path = os.path.join(self.plugin.builddir, parse_relpath)

            with contextlib.suppress(errors.MissingMetadataFileError):
                metadata.update(
                    extract_metadata(self.name, parse_relpath, self.plugin.installdir)
                )
                found_path = os.path.join(self.plugin.installdir, parse_relpath)

            if found_path is not None:
                metadata_files.append(found_path)
            else:
                # If this metadata file is not found in either build or
                # install, check the pull state to make sure it was found
                # there. If not, we need to let the user know.
                state = self.get_pull_state()
                if not state or parse_relpath not in state.extracted_metadata["files"]:
                    raise errors.MissingMetadataFileError(self.name, parse_relpath)

        self.mark_done(
            steps.BUILD,
            states.BuildState(
                property_names=build_properties,
                part_properties=self._part_properties,
                project=self._project_options,
                plugin_assets=plugin_manifest,
                machine_assets=machine_manifest,
                metadata=metadata,
                metadata_files=metadata_files,
                scriptlet_metadata=self._scriptlet_metadata[steps.BUILD],
            ),
        )

    def _get_machine_manifest(self):
        # Use subprocess directly here. common.run_output will use binaries out
        # of the snap, and we want to use the one on the host.
        try:
            output = subprocess.check_output(
                [
                    "uname",
                    "--kernel-name",
                    "--kernel-release",
                    "--kernel-version",
                    "--machine",
                    "--processor",
                    "--hardware-platform",
                    "--operating-system",
                ]
            )
        except subprocess.CalledProcessError as e:
            logger.warning(
                "'uname' exited with code {}: unable to record machine "
                "manifest".format(e.returncode)
            )
            return {}

        try:
            uname = output.decode(sys.getfilesystemencoding()).strip()
        except UnicodeEncodeError:
            logger.warning("Could not decode output for 'uname' correctly")
            uname = output.decode("latin-1", "surrogateescape").strip()

        return {
            "uname": uname,
            "installed-packages": repo.Repo.get_installed_packages(),
            "installed-snaps": repo.snaps.get_installed_snaps(),
        }

    def clean_build(self):
        if self.is_clean(steps.BUILD):
            return

        if os.path.exists(self.plugin.build_basedir):
            shutil.rmtree(self.plugin.build_basedir)

        if os.path.exists(self.plugin.installdir):
            shutil.rmtree(self.plugin.installdir)

        self.plugin.clean_build()
        self.mark_cleaned(steps.BUILD)

    def migratable_fileset_for(self, step):
        plugin_fileset = self.plugin.snap_fileset()
        fileset = self._get_fileset(step.name).copy()
        includes = _get_includes(fileset)
        # If we're priming and we don't have an explicit set of files to prime
        # include the files from the stage step
        if step == steps.PRIME and (fileset == ["*"] or len(includes) == 0):
            stage_fileset = self._get_fileset(steps.STAGE.name).copy()
            fileset = _combine_filesets(stage_fileset, fileset)

        fileset.extend(plugin_fileset)

        return _migratable_filesets(fileset, self.plugin.installdir)

    def _get_fileset(self, option, default=None):
        if default is None:
            default = ["*"]

        fileset = getattr(self.plugin.options, option, default)
        return fileset if fileset else default

    def _organize(self, *, overwrite=False):
        fileset = self._get_fileset("organize", {})

        _organize_filesets(self.name, fileset.copy(), self.plugin.installdir, overwrite)

    def stage(self, force=False):
        self._do_runner_step(steps.STAGE)

        # Only mark this step done if _do_stage() didn't run, in which case
        # we have no directories or files to track.
        if self.is_clean(steps.STAGE):
            self.mark_stage_done(set(), set())

    def _do_stage(self):
        snap_files, snap_dirs = self.migratable_fileset_for(steps.STAGE)

        def fixup_func(file_path):
            if os.path.islink(file_path):
                return
            if not file_path.endswith(".pc"):
                return
            repo.fix_pkg_config(self.stagedir, file_path, self.plugin.installdir)

        _migrate_files(
            snap_files,
            snap_dirs,
            self.plugin.installdir,
            self.stagedir,
            fixup_func=fixup_func,
        )
        # TODO once `snappy try` is in place we will need to copy
        # dependencies here too

        self.mark_stage_done(snap_files, snap_dirs)

    def mark_stage_done(self, snap_files, snap_dirs):
        self.mark_done(
            steps.STAGE,
            states.StageState(
                snap_files,
                snap_dirs,
                self._part_properties,
                self._project_options,
                self._scriptlet_metadata[steps.STAGE],
            ),
        )

    def clean_stage(self, project_staged_state):
        if self.is_clean(steps.STAGE):
            return

        state = states.get_state(self.plugin.statedir, steps.STAGE)

        try:
            self._clean_shared_area(self.stagedir, state, project_staged_state)
        except AttributeError:
            raise errors.MissingStateCleanError(steps.STAGE)

        self.mark_cleaned(steps.STAGE)

    def prime(self, force=False) -> None:
        self._do_runner_step(steps.PRIME)

        # Only mark this step done if _do_prime() didn't run, in which case
        # we have no files, directories, or dependency paths to track.
        if self.is_clean(steps.PRIME):
            self.mark_prime_done(set(), set(), set())

    def _do_prime(self) -> None:
        snap_files, snap_dirs = self.migratable_fileset_for(steps.PRIME)
        _migrate_files(snap_files, snap_dirs, self.stagedir, self.primedir)

        if self._snap_type == "app":
            dependency_paths = self._handle_elf(snap_files)
        else:
            dependency_paths = set()

        self.mark_prime_done(snap_files, snap_dirs, dependency_paths)

    def _handle_elf(self, snap_files: Sequence[str]) -> Set[str]:
        elf_files = elf.get_elf_files(self.primedir, snap_files)
        all_dependencies = set()
        core_path = common.get_core_path(self._base)

        # Clear the cache of all libs that aren't already in the primedir
        self._soname_cache.reset_except_root(self.primedir)
        for elf_file in elf_files:
            all_dependencies.update(
                elf_file.load_dependencies(
                    root_path=self.primedir,
                    core_base_path=core_path,
                    soname_cache=self._soname_cache,
                )
            )

        dependency_paths = self._handle_dependencies(all_dependencies)

        if not self._build_attributes.keep_execstack():
            clear_execstack(elf_files=elf_files)

        if self._build_attributes.no_patchelf():
            logger.warning(
                "The primed files for part {!r} will not be verified for "
                "correctness or patched: build-attributes: [no-patchelf] "
                "is set.".format(self.name)
            )
        else:
            part_patcher = PartPatcher(
                elf_files=elf_files,
                plugin=self.plugin,
                project=self._project_options,
                confinement=self._confinement,
                core_base=self._base,
                snap_base_path=self._snap_base_path,
                stagedir=self.stagedir,
                primedir=self.primedir,
                stage_packages=self._part_properties.get("stage-packages", []),
            )
            part_patcher.patch()

        return dependency_paths

    def mark_prime_done(self, snap_files, snap_dirs, dependency_paths):
        self.mark_done(
            steps.PRIME,
            states.PrimeState(
                snap_files,
                snap_dirs,
                dependency_paths,
                self._part_properties,
                self._project_options,
                self._scriptlet_metadata[steps.PRIME],
            ),
        )

    def clean_prime(self, project_primed_state, hint=""):
        if self.is_clean(steps.PRIME):
            return

        state = self.get_prime_state()

        try:
            self._clean_shared_area(self.primedir, state, project_primed_state)
        except AttributeError:
            raise errors.MissingStateCleanError(steps.PRIME)

        self.mark_cleaned(steps.PRIME)

    def _clean_shared_area(self, shared_directory, part_state, project_state):
        primed_files = part_state.files
        primed_directories = part_state.directories

        # We want to make sure we don't remove a file or directory that's
        # being used by another part. So we'll examine the state for all parts
        # in the project and leave any files or directories found to be in
        # common.
        for other_name, other_state in project_state.items():
            if other_state and (other_name != self.name):
                primed_files -= other_state.files
                primed_directories -= other_state.directories

        # Finally, clean the files and directories that are specific to this
        # part.
        _clean_migrated_files(primed_files, primed_directories, shared_directory)

    def _handle_dependencies(self, all_dependencies: Set[str]):
        # Split the necessary dependencies into their corresponding location.
        # We'll only track the part and staged dependencies, since they should have
        # already been primed by other means, and migrating them again could
        # potentially override the `stage` or `snap` filtering.
        (in_part, staged, primed, system) = _split_dependencies(
            all_dependencies, self.plugin.installdir, self.stagedir, self.primedir
        )
        part_dependency_paths = {os.path.dirname(d) for d in in_part}
        staged_dependency_paths = {os.path.dirname(d) for d in staged}
        dependency_paths = part_dependency_paths | staged_dependency_paths

        resolver = MissingDependencyResolver(elf_files=system)
        resolver.print_resolutions(
            part_name=self.name,
            stage_packages_exist=self._part_properties.get("stage-packages"),
            echoer=logger,
        )

        return dependency_paths

    def get_primed_dependency_paths(self):
        dependency_paths = set()
        state = self.get_prime_state()
        if state:
            for path in state.dependency_paths:
                dependency_paths.add(os.path.join(self.primedir, path.lstrip("/")))

        return dependency_paths

    def env(self, root):
        return self.plugin.env(root)

    def clean(self, project_staged_state=None, project_primed_state=None, step=None):
        if not project_staged_state:
            project_staged_state = {}

        if not project_primed_state:
            project_primed_state = {}

        try:
            self._clean_steps(project_staged_state, project_primed_state, step)
        except errors.MissingStateCleanError:
            # If one of the step cleaning rules is missing state, it must be
            # running on the output of an old Snapcraft. In that case, if we
            # were specifically asked to clean that step we need to fail.
            # Otherwise, just clean like the old Snapcraft did, and blow away
            # the entire part directory.
            if step:
                raise

            logger.info("Cleaning up for part {!r}".format(self.name))
            if os.path.exists(self.plugin.partdir):
                shutil.rmtree(self.plugin.partdir)

        # Remove the part directory if it's completely empty (i.e. all steps
        # have been cleaned).
        if os.path.exists(self.plugin.partdir) and not os.listdir(self.plugin.partdir):
            os.rmdir(self.plugin.partdir)

    def _clean_steps(self, project_staged_state, project_primed_state, step=None):
        if step:
            if step not in steps.STEPS:
                raise RuntimeError(
                    "{!r} is not a valid step for part {!r}".format(step, self.name)
                )

        if not step or step <= steps.PRIME:
            self.clean_prime(project_primed_state)

        if not step or step <= steps.STAGE:
            self.clean_stage(project_staged_state)

        if not step or step <= steps.BUILD:
            self.clean_build()

        if not step or step <= steps.PULL:
            self.clean_pull()


def _split_dependencies(dependencies, installdir, stagedir, primedir):
    """Split dependencies into their corresponding location.

    Return a tuple of sets for each location.
    """

    part_dependencies = set()
    staged_dependencies = set()
    primed_dependencies = set()
    system_dependencies = set()

    for file_path in dependencies:
        if file_path.startswith(installdir):
            part_dependencies.add(os.path.relpath(file_path, installdir))
        elif file_path.startswith(stagedir):
            staged_dependencies.add(os.path.relpath(file_path, stagedir))
        elif file_path.startswith(primedir):
            primed_dependencies.add(os.path.relpath(file_path, primedir))
        else:
            file_path = file_path.lstrip("/")

            # This was a dependency that was resolved to be on the system.
            # However, it's possible that this library is actually included in
            # the snap and we just missed it because it's in a non-standard
            # path. Let's make sure it isn't already in the part, stage dir, or
            # prime dir. If so, add it to that set.
            if os.path.exists(os.path.join(installdir, file_path)):
                part_dependencies.add(file_path)
            elif os.path.exists(os.path.join(stagedir, file_path)):
                staged_dependencies.add(file_path)
            elif os.path.exists(os.path.join(primedir, file_path)):
                primed_dependencies.add(file_path)
            else:
                system_dependencies.add(file_path)

    return (
        part_dependencies,
        staged_dependencies,
        primed_dependencies,
        system_dependencies,
    )


def _expand_part_properties(part_properties, part_schema):
    """Returns properties with all part schema properties included.

    Any schema properties not set will contain their default value as defined
    in the schema itself.
    """

    # First make a deep copy of the part schema. It contains nested mutables,
    # and we'd rather not change them.
    part_schema = copy.deepcopy(part_schema)

    # Come up with a dictionary of part schema properties and their default
    # values as defined in the schema.
    properties = {}
    for schema_property, subschema in part_schema.items():
        properties[schema_property] = subschema.get("default")

    # Now expand (overwriting if necessary) the default schema properties with
    # the ones from the actual part.
    properties.update(part_properties)

    return properties


def _migratable_filesets(fileset, srcdir):
    includes, excludes = _get_file_list(fileset)

    include_files = _generate_include_set(srcdir, includes)
    exclude_files, exclude_dirs = _generate_exclude_set(srcdir, excludes)

    # Chop files, including whole trees if any dirs are mentioned.
    snap_files = include_files - exclude_files
    for exclude_dir in exclude_dirs:
        snap_files = set([x for x in snap_files if not x.startswith(exclude_dir + "/")])

    # Separate dirs from files.
    snap_dirs = set(
        [
            x
            for x in snap_files
            if os.path.isdir(os.path.join(srcdir, x))
            and not os.path.islink(os.path.join(srcdir, x))
        ]
    )

    # Remove snap_dirs from snap_files.
    snap_files = snap_files - snap_dirs

    # Include (resolved) parent directories for each selected file.
    for snap_file in snap_files:
        snap_file = file_utils.get_resolved_relative_path(snap_file, srcdir)
        dirname = os.path.dirname(snap_file)
        while dirname:
            snap_dirs.add(dirname)
            dirname = os.path.dirname(dirname)

    # Resolve parent paths for dirs and files.
    resolved_snap_dirs = set()
    for snap_dir in snap_dirs:
        snap_dir = file_utils.get_resolved_relative_path(snap_dir, srcdir)
        resolved_snap_dirs.add(snap_dir)

    resolved_snap_files = set()
    for snap_file in snap_files:
        snap_file = file_utils.get_resolved_relative_path(snap_file, srcdir)
        resolved_snap_files.add(snap_file)

    return resolved_snap_files, resolved_snap_dirs


def _migrate_files(
    snap_files,
    snap_dirs,
    srcdir,
    dstdir,
    missing_ok=False,
    follow_symlinks=False,
    fixup_func=lambda *args: None,
):
    for snap_dir in sorted(snap_dirs):
        src = os.path.join(srcdir, snap_dir)
        dst = os.path.join(dstdir, snap_dir)

        snapcraft.file_utils.create_similar_directory(src, dst)

    for snap_file in sorted(snap_files):
        src = os.path.join(srcdir, snap_file)
        dst = os.path.join(dstdir, snap_file)

        if missing_ok and not os.path.exists(src):
            continue

        # If the file is already here and it's a symlink, leave it alone.
        if os.path.islink(dst):
            continue

        # Otherwise, remove and re-link it.
        if os.path.exists(dst):
            os.remove(dst)

        if src.endswith(".pc"):
            shutil.copy2(src, dst, follow_symlinks=follow_symlinks)
        else:
            file_utils.link_or_copy(src, dst, follow_symlinks=follow_symlinks)

        fixup_func(dst)


def _organize_filesets(part_name, fileset, base_dir, overwrite):
    for key in sorted(fileset, key=lambda x: ["*" in x, x]):
        src = os.path.join(base_dir, key)
        # Remove the leading slash if there so os.path.join
        # actually joins
        dst = os.path.join(base_dir, fileset[key].lstrip("/"))

        sources = iglob(src, recursive=True)

        # Keep track of the number of glob expansions so we can properly error if more
        # than one tries to organize to the same file
        src_count = 0
        for src in sources:
            src_count += 1

            if os.path.isdir(src) and "*" not in key:
                file_utils.link_or_copy_tree(src, dst)
                # TODO create alternate organization location to avoid
                # deletions.
                shutil.rmtree(src)
                continue
            elif os.path.isfile(dst):
                if overwrite and src_count <= 1:
                    with contextlib.suppress(FileNotFoundError):
                        os.remove(dst)
                elif src_count > 1:
                    raise errors.SnapcraftOrganizeError(
                        part_name,
                        "multiple files to be organized into {!r}. If this is supposed "
                        "to be a directory, end it with a forward slash.".format(
                            os.path.relpath(dst, base_dir)
                        ),
                    )
                else:
                    raise errors.SnapcraftOrganizeError(
                        part_name,
                        "trying to organize file {key!r} to {dst!r}, but {dst!r} "
                        "already exists".format(
                            key=key, dst=os.path.relpath(dst, base_dir)
                        ),
                    )
            if os.path.isdir(dst) and overwrite:
                real_dst = os.path.join(dst, os.path.basename(src))
                if os.path.isdir(real_dst):
                    shutil.rmtree(real_dst)
                else:
                    with contextlib.suppress(FileNotFoundError):
                        os.remove(real_dst)
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            shutil.move(src, dst)


def _clean_migrated_files(snap_files, snap_dirs, directory):
    for snap_file in snap_files:
        try:
            os.remove(os.path.join(directory, snap_file))
        except FileNotFoundError:
            logger.warning(
                "Attempted to remove file {name!r}, but it didn't exist. "
                "Skipping...".format(name=snap_file)
            )

    # snap_dirs may not be ordered so that subdirectories come before
    # parents, and we want to be able to remove directories if possible, so
    # we'll sort them in reverse here to get subdirectories before parents.
    snap_dirs = sorted(snap_dirs, reverse=True)

    for snap_dir in snap_dirs:
        migrated_directory = os.path.join(directory, snap_dir)
        try:
            if not os.listdir(migrated_directory):
                os.rmdir(migrated_directory)
        except FileNotFoundError:
            logger.warning(
                "Attempted to remove directory {name!r}, but it didn't exist. "
                "Skipping...".format(name=snap_dir)
            )


def _get_file_list(stage_set):
    includes = []
    excludes = []

    for item in stage_set:
        if item.startswith("-"):
            excludes.append(item[1:])
        elif item.startswith("\\"):
            includes.append(item[1:])
        else:
            includes.append(item)

    _validate_relative_paths(includes + excludes)

    includes = includes or ["*"]

    return includes, excludes


def _generate_include_set(directory, includes):
    include_files = set()
    for include in includes:
        if "*" in include:
            pattern = os.path.join(directory, include)
            matches = iglob(pattern, recursive=True)
            include_files |= set(matches)
        else:
            include_files |= set([os.path.join(directory, include)])

    include_dirs = [x for x in include_files if os.path.isdir(x)]
    include_files = set([os.path.relpath(x, directory) for x in include_files])

    # Expand includeFiles, so that an exclude like '*/*.so' will still match
    # files from an include like 'lib'
    for include_dir in include_dirs:
        for root, dirs, files in os.walk(include_dir):
            include_files |= set(
                [os.path.relpath(os.path.join(root, d), directory) for d in dirs]
            )
            include_files |= set(
                [os.path.relpath(os.path.join(root, f), directory) for f in files]
            )

    return include_files


def _generate_exclude_set(directory, excludes):
    exclude_files = set()

    for exclude in excludes:
        pattern = os.path.join(directory, exclude)
        matches = iglob(pattern, recursive=True)
        exclude_files |= set(matches)

    exclude_dirs = [
        os.path.relpath(x, directory) for x in exclude_files if os.path.isdir(x)
    ]
    exclude_files = set([os.path.relpath(x, directory) for x in exclude_files])

    return exclude_files, exclude_dirs


def _validate_relative_paths(files):
    for d in files:
        if os.path.isabs(d):
            raise errors.PluginError('path "{}" must be relative'.format(d))


def _file_collides(file_this, file_other):
    if not file_this.endswith(".pc"):
        return not filecmp.cmp(file_this, file_other, shallow=False)

    pc_file_1 = open(file_this)
    pc_file_2 = open(file_other)

    try:
        for lines in zip(pc_file_1, pc_file_2):
            for line in zip(lines[0].split("\n"), lines[1].split("\n")):
                if line[0].startswith("prefix="):
                    continue
                if line[0] != line[1]:
                    return True
    except Exception as e:
        raise e from e
    finally:
        pc_file_1.close()
        pc_file_2.close()
    return False


def check_for_collisions(parts):
    """Raises a SnapcraftPartConflictError if conflicts are found."""
    parts_files = {}
    for part in parts:
        # Gather our own files up
        part_files, part_directories = part.migratable_fileset_for(steps.STAGE)
        part_contents = part_files | part_directories

        # Scan previous parts for collisions
        for other_part_name in parts_files:
            common = part_contents & parts_files[other_part_name]["files"]
            conflict_files = []
            for f in common:
                this = os.path.join(part.plugin.installdir, f)
                other = os.path.join(parts_files[other_part_name]["installdir"], f)

                if _paths_collide(this, other):
                    conflict_files.append(f)

            if conflict_files:
                raise errors.SnapcraftPartConflictError(
                    other_part_name=other_part_name,
                    part_name=part.name,
                    conflict_files=conflict_files,
                )

        # And add our files to the list
        parts_files[part.name] = {
            "files": part_contents,
            "installdir": part.plugin.installdir,
        }


def _paths_collide(path1: str, path2: str) -> bool:
    if not (os.path.lexists(path1) and os.path.lexists(path2)):
        return False

    path1_is_dir = os.path.isdir(path1)
    path2_is_dir = os.path.isdir(path2)
    path1_is_link = os.path.islink(path1)
    path2_is_link = os.path.islink(path2)

    # Paths collide if they're both symlinks, but pointing to different places
    if path1_is_link and path2_is_link:
        return os.readlink(path1) != os.readlink(path2)

    # Paths collide if one is a symlink, but not the other
    elif path1_is_link or path2_is_link:
        return True

    # Paths collide if one is a directory, but not the other
    elif path1_is_dir != path2_is_dir:
        return True

    # Paths collide if neither path is a directory, and the files have
    # different contents
    elif not (path1_is_dir and path2_is_dir) and _file_collides(path1, path2):
        return True

    # Otherwise, paths do not conflict
    else:
        return False


def _get_includes(fileset):
    return [x for x in fileset if x[0] != "-"]


def _get_excludes(fileset):
    return [x[1:] for x in fileset if x[0] == "-"]


def _combine_filesets(starting_fileset, modifying_fileset):
    """
    Combine filesets if modifying_fileset is an explicit or implicit
    wildcard.
    """

    starting_excludes = set(_get_excludes(starting_fileset))
    modifying_includes = set(_get_includes(modifying_fileset))

    contradicting_fileset = set.intersection(starting_excludes, modifying_includes)

    if contradicting_fileset:
        raise errors.PrimeFileConflictError(fileset=contradicting_fileset)

    to_combine = False
    # combine if starting_fileset has a wildcard
    # XXX: should this only be a single wildcard and possibly excludes?
    if "*" in modifying_fileset:
        to_combine = True
        modifying_fileset.remove("*")

    # combine if modifying_fileset is only excludes
    if set([x[0] for x in modifying_fileset]) == set("-"):
        to_combine = True

    if to_combine:
        return list(set(starting_fileset + modifying_fileset))
    else:
        return modifying_fileset


def _list_of_dicts_to_env(l: List[Dict[str, str]]) -> List[str]:
    env = []  # type: List[str]

    # We're iterating anyway, but thanks to the schema validation, we can rest assured
    # that each dict only has one key/value pair.
    for d in l:
        for key, value in d.items():
            env.append('{}="{}"'.format(key, value))

    return env



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Create snaps from conda packages.

This plugin uses the common plugin keywords as well as those for 'sources'.
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

In the cases where dumping the content needs some mangling or organizing
one would take advantage of the core functionalities available to plugins
such as: `filesets`, `stage`, `snap` and `organize`.
"""

import os
import shutil
import subprocess

import snapcraft
from snapcraft.internal import errors, sources


# Supported versions https://repo.anaconda.com/miniconda/
_MINICONDA_URL = {
    "4.6.14": dict(
        source="https://repo.anaconda.com/miniconda/Miniconda3-4.6.14-Linux-x86_64.sh",
        checksum="md5/718259965f234088d785cad1fbd7de03",
    ),
    "latest": dict(
        source="https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh",
        checksum=None,
    ),
}

_MINICONDA_UNKNOWN_VERSION_URL = dict(
    source="https://repo.anaconda.com/miniconda/Miniconda3-{}-Linux-x86_64.sh",
    checksum=None,
)


class MinicondaBadArchitectureError(snapcraft.internal.errors.SnapcraftError):

    fmt = (
        "Failed to fetch miniconda: "
        "The architecture {architecture!r} is not supported."
    )

    def __init__(self, *, architecture: str) -> None:
        super().__init__(architecture=architecture)


class CondaPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["conda-packages"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["conda-python-version"] = {"type": "string", "default": ""}
        schema["properties"]["conda-miniconda-version"] = {
            "type": "string",
            "default": "latest",
        }
        schema["required"] = ["conda-packages"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["conda-miniconda-version"]

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["conda-python-version", "conda-packages"]

    def __init__(self, name, options, project) -> None:
        super().__init__(name, options, project)
        if project.info.get_build_base() not in ("core", "core16", "core18"):
            raise errors.PluginBaseError(
                part_name=self.name, base=project.info.get_build_base()
            )

        self._conda_home = os.path.join(self.partdir, "miniconda")
        self._miniconda_script = os.path.join(self.partdir, "miniconda.sh")

    def _get_miniconda_source(self) -> sources.Script:
        miniconda_url = _MINICONDA_URL.get(self.options.conda_miniconda_version)
        if not miniconda_url:
            miniconda_url = _MINICONDA_UNKNOWN_VERSION_URL
            miniconda_url["source"] = miniconda_url["source"].format(
                self.options.conda_miniconda_version
            )

        return sources.Script(
            source=miniconda_url["source"],
            source_dir=self.partdir,
            source_checksum=miniconda_url["checksum"],
        )

    def pull(self):
        if self.project.deb_arch != "amd64":
            raise MinicondaBadArchitectureError(architecture=self.project.deb_arch)
        miniconda_source_script = self._get_miniconda_source()
        miniconda_source_script.download(filepath=self._miniconda_script)
        os.chmod(self._miniconda_script, mode=0o755)

    def clean_pull(self):
        super().clean_pull()

        # Remove conda_home (if present)
        if os.path.exists(self._conda_home):
            shutil.rmtree(self._conda_home)

    def build(self):
        super().build()

        conda_cmd = os.path.join(self._conda_home, "bin", "conda")

        # -b: batch mode (agrees to license and does not inject into bash profile)
        # -f: force
        # -p: prefix
        if not os.path.exists(conda_cmd):
            subprocess.check_call([self._miniconda_script, "-bfp", self._conda_home])

        deploy_cmd = [conda_cmd, "create", "--prefix", self.installdir, "--yes"]
        if self.options.conda_python_version:
            deploy_cmd.append("python={}".format(self.options.conda_python_version))

        # conda needs to rewrite the prefixes in the python shebangs and binaries
        conda_target_prefix = os.path.join(
            os.path.sep, "snap", self.project.info.name, "current"
        )

        subprocess.check_call(
            deploy_cmd + self.options.conda_packages,
            env=dict(CONDA_TARGET_PREFIX_OVERRIDE=conda_target_prefix),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import os
import time
from subprocess import check_call, CalledProcessError, Popen, PIPE, STDOUT

from progressbar import AnimatedMarker, ProgressBar

from . import errors
from snapcraft import file_utils, yaml_utils
from snapcraft.internal import common
from snapcraft.internal.indicators import is_dumb_terminal


_SNAP_PATH = os.path.join(os.path.sep, "snap", "core", "current", "usr", "bin", "snap")


logger = logging.getLogger(__name__)


def _snap_data_from_dir(directory):
    with open(os.path.join(directory, "meta", "snap.yaml")) as f:
        snap = yaml_utils.load(f)

    return {
        "name": snap["name"],
        "version": snap["version"],
        "arch": snap.get("architectures", []),
        "type": snap.get("type", ""),
        "license": snap.get("license"),
    }


def pack(directory, output=None):
    mksquashfs_path = file_utils.get_tool_path("mksquashfs")

    snap = _snap_data_from_dir(directory)
    # Verify that the snap command is available and use it to
    # validate the layout.
    # The snap command will most likely be found as it lives in
    # core and the snapcraft snap lives on top of it (on the side
    # rather).
    if os.path.exists(_SNAP_PATH):
        _run_snap_pack_verification(directory=directory)
    elif snap.get("license"):
        logger.warning(
            "Could not find {!r}, validation of the license string will only "
            "take place once pushed to the store.".format(_SNAP_PATH)
        )

    output_snap_name = output or common.format_snap_name(snap)
    # If a .snap-build exists at this point, when we are about to override
    # the snap blob, it is stale. We rename it so user have a chance to
    # recover accidentally lost assertions.
    snap_build = output_snap_name + "-build"
    if os.path.isfile(snap_build):
        _new = "{}.{}".format(snap_build, int(time.time()))
        logger.warning("Renaming stale build assertion to {}".format(_new))
        os.rename(snap_build, _new)

    _run_mksquashfs(
        mksquashfs_path,
        directory=directory,
        snap_name=snap["name"],
        snap_type=snap["type"],
        output_snap_name=output_snap_name,
    )

    return output_snap_name


def _run_snap_pack_verification(*, directory: str) -> None:
    try:
        check_call([_SNAP_PATH, "pack", "--check-skeleton", directory])
    except CalledProcessError:
        raise errors.PackVerificationError()


def _run_mksquashfs(
    mksquashfs_command, *, directory, snap_name, snap_type, output_snap_name
):
    # These options need to match the review tools:
    # http://bazaar.launchpad.net/~click-reviewers/click-reviewers-tools/trunk/view/head:/clickreviews/common.py#L38
    mksquashfs_args = ["-noappend", "-comp", "xz", "-no-xattrs", "-no-fragments"]
    if snap_type != "base":
        mksquashfs_args.append("-all-root")

    complete_command = [
        mksquashfs_command,
        directory,
        output_snap_name,
    ] + mksquashfs_args

    with Popen(complete_command, stdout=PIPE, stderr=STDOUT) as proc:
        ret = None
        if is_dumb_terminal():
            logger.info("Snapping {!r} ...".format(snap_name))
            ret = proc.wait()
        else:
            message = "\033[0;32m\rSnapping {!r}\033[0;32m ".format(snap_name)
            progress_indicator = ProgressBar(
                widgets=[message, AnimatedMarker()], maxval=7
            )
            progress_indicator.start()

            ret = proc.poll()
            count = 0

            while ret is None:
                if count >= 7:
                    progress_indicator.start()
                    count = 0
                progress_indicator.update(count)
                count += 1
                time.sleep(0.2)
                ret = proc.poll()
        print("")
        if ret != 0:
            logger.error(proc.stdout.read().decode("utf-8"))
            raise RuntimeError("Failed to create snap {!r}".format(output_snap_name))

        logger.debug(proc.stdout.read().decode("utf-8"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import os
import json
from collections import OrderedDict
from typing import Any, Dict, TYPE_CHECKING

import snapcraft
from snapcraft.internal import errors, os_release, steps
from snapcraft.internal.states import GlobalState, get_state

if TYPE_CHECKING:
    from snapcraft.project import Project


def annotate_snapcraft(project: "Project", data: Dict[str, Any]) -> Dict[str, Any]:
    manifest = OrderedDict()  # type: Dict[str, Any]
    manifest["snapcraft-version"] = snapcraft._get_version()
    manifest["snapcraft-started-at"] = project._get_start_time().isoformat() + "Z"

    release = os_release.OsRelease()
    with contextlib.suppress(errors.OsReleaseIdError):
        manifest["snapcraft-os-release-id"] = release.id()
    with contextlib.suppress(errors.OsReleaseVersionIdError):
        manifest["snapcraft-os-release-version-id"] = release.version_id()

    for k, v in data.items():
        manifest[k] = v
    image_info = os.environ.get("SNAPCRAFT_IMAGE_INFO")
    if image_info:
        try:
            image_info_dict = json.loads(image_info)
        except json.decoder.JSONDecodeError as exception:
            raise errors.InvalidContainerImageInfoError(image_info) from exception
        manifest["image-info"] = image_info_dict

    global_state = GlobalState.load(filepath=project._get_global_state_file_path())
    manifest["build-packages"] = global_state.get_build_packages()
    manifest["build-snaps"] = global_state.get_build_snaps()

    for part in data["parts"]:
        state_dir = os.path.join(project.parts_dir, part, "state")
        pull_state = get_state(state_dir, steps.PULL)
        manifest["parts"][part]["build-packages"] = pull_state.assets.get(
            "build-packages", []
        )
        manifest["parts"][part]["stage-packages"] = pull_state.assets.get(
            "stage-packages", []
        )
        source_details = pull_state.assets.get("source-details", {})
        if source_details:
            manifest["parts"][part].update(source_details)
        build_state = get_state(state_dir, steps.BUILD)
        manifest["parts"][part].update(build_state.assets)
    return manifest



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
from typing import List, Tuple

from snapcraft.internal import pluginhandler, steps
import snapcraft.internal.errors
from . import errors


def latest_step(
    parts: List[pluginhandler.PluginHandler],
) -> Tuple[pluginhandler.PluginHandler, steps.Step, int]:
    """Determine and return the latest step that was run.

    :param list parts: List of all parts.
    :returns: Tuple of (part, step, timestamp).
    """
    latest_part = None
    latest_step = None
    latest_timestamp = 0
    for part in parts:
        with contextlib.suppress(snapcraft.internal.errors.NoLatestStepError):
            step = part.latest_step()
            timestamp = part.step_timestamp(step)
            if latest_timestamp < timestamp:
                latest_part = part
                latest_step = step
                latest_timestamp = timestamp

    if not latest_part:
        raise errors.NoStepsRunError()

    return (latest_part, latest_step, latest_timestamp)
